"""PvE Views"""

# Standard Library
import json

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.core.handlers.wsgi import WSGIRequest
from django.shortcuts import redirect, render
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.models import EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.api.helpers import get_all_corporations_from_alliance, get_alliance
from ledger.helpers.alliance import AllianceData
from ledger.helpers.core import LedgerEntity, add_info_to_context
from ledger.models.corporationaudit import CorporationWalletJournalEntry

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("ledger.advanced_access")
def alliance_ledger_index(request):
    """Alliance Ledger Index View"""
    context = {}
    context = add_info_to_context(request, context)
    alliance_id = request.user.profile.main_character.alliance_id
    if not alliance_id:
        alliance_id = 0

    return redirect("ledger:alliance_ledger", alliance_id=alliance_id)


@login_required
@permission_required("ledger.advanced_access")
def alliance_ledger(request, alliance_id, year=None, month=None, day=None):
    """
    Alliance Ledger
    """
    perms, alliance = get_alliance(request, alliance_id)

    context = {
        "title": "Alliance Ledger",
        "alliance_id": alliance_id,
        "disabled": True,
    }

    # pylint: disable=duplicate-code
    if perms is False:
        msg = _("Permission Denied")
        messages.error(request, msg)
        return render(
            request, "ledger/allyledger/alliance_ledger.html", context=context
        )
    # pylint: disable=duplicate-code
    if perms is None:
        msg = _("Alliance not found.")
        messages.info(request, msg)
        return render(
            request, "ledger/allyledger/alliance_ledger.html", context=context
        )

    alliance_data = AllianceData(
        request=request, alliance=alliance, year=year, month=month, day=day
    )

    # Create the Alliance ledger data
    ledger = alliance_data.generate_ledger_data()

    context = {
        "title": f"Alliance Ledger - {alliance.alliance_name}",
        "alliance_id": alliance_id,
        "billboard": json.dumps(alliance_data.billboard.dict.asdict()),
        "ledger": ledger,
        "years": CorporationWalletJournalEntry.objects.filter(
            division__corporation__corporation__corporation_id__in=alliance_data.corporations
        )
        .values_list("date__year", flat=True)
        .distinct()
        .order_by("-date__year"),
        "totals": alliance_data.calculate_totals(ledger),
        "view": alliance_data.create_view_data(
            viewname="alliance_details",
            alliance_id=alliance_id,
            entity_id=alliance_id,
            section="summary",
        ),
    }
    # Add additional information to the context
    context = add_info_to_context(request, context)

    return render(request, "ledger/allyledger/alliance_ledger.html", context=context)


# pylint: disable=too-many-positional-arguments
@login_required
@permission_required("ledger.advanced_access")
def alliance_details(
    request: WSGIRequest,
    alliance_id=None,
    entity_id=None,
    year=None,
    month=None,
    day=None,
    section=None,
):
    """
    Alliance Details
    """
    perms, alliance = get_alliance(request, alliance_id=alliance_id)

    # pylint: disable=duplicate-code
    if perms is False:
        msg = _("Permission Denied")
        return render(
            request,
            "ledger/partials/information/view_character_content.html",
            {
                "error": msg,
                "alliance_id": alliance_id,
            },
        )
    # pylint: disable=duplicate-code
    if perms is None:
        msg = _("Alliance not found.")
        return render(
            request,
            "ledger/partials/information/view_character_content.html",
            {
                "error": msg,
                "alliance_id": alliance_id,
            },
        )

    alliance_data = AllianceData(
        request=request,
        alliance=alliance,
        year=year,
        month=month,
        day=day,
        section=section,
    )

    # Create the Entity for the ledger
    entity = LedgerEntity(
        entity_id=entity_id,
        alliance_obj=alliance if entity_id == alliance.alliance_id else None,
    )

    journal = alliance_data.filter_entity_journal(entity=entity)
    amounts = alliance_data._create_corporation_details(journal=journal, entity=entity)
    details = alliance_data._add_average_details(request, amounts, day)

    context = {
        "title": f"Alliance Details - {alliance.alliance_name}",
        "type": "alliance",
        "character": details,
        "information": f"Alliance Details - {alliance_data.get_details_title}",
    }
    context = add_info_to_context(request, context)
    # pylint: disable=duplicate-code
    return render(
        request,
        "ledger/partials/information/view_character_content.html",
        context=context,
    )


@login_required
@permission_required("ledger.advanced_access")
def alliance_overview(request):
    """
    Alliance Overview
    """
    context = {
        "title": "Alliance Overview",
    }
    context = add_info_to_context(request, context)
    return render(
        request, "ledger/allyledger/admin/alliance_overview.html", context=context
    )


@login_required
@permission_required("ledger.manage_access")
def alliance_administration(request, alliance_id):
    """
    Alliance Administration
    """
    perm, alliance = get_alliance(request, alliance_id)

    if perm is False:
        msg = _("Permission Denied")
        messages.error(request, msg)
        return redirect("ledger:alliance_ledger_index")
    if perm is None:
        msg = _("Alliance not found.")
        messages.info(request, msg)
        return redirect("ledger:alliance_ledger_index")

    corporations = get_all_corporations_from_alliance(request, alliance.alliance_id)[1]
    all_corporations = EveCorporationInfo.objects.filter(
        alliance__alliance_id=alliance_id
    ).order_by("corporation_name")

    corp_audit_ids = corporations.values_list("corporation__corporation_id", flat=True)
    missing_corporations = all_corporations.exclude(
        corporation_id__in=corp_audit_ids
    ).order_by("corporation_name")

    context = {
        "alliance_id": alliance_id,
        "title": "Alliance Administration",
        "alliance": alliance,
        "corporations": corporations,
        "missing_corporations": missing_corporations,
    }
    context = add_info_to_context(request, context)
    return render(
        request,
        "ledger/allyledger/admin/alliance_administration.html",
        context=context,
    )



"""
Planetary Audit
"""

# Standard Library
from http import HTTPStatus

# Django
from django.contrib.auth.decorators import login_required, permission_required
from django.db.models import Q
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.utils.translation import gettext_lazy as _
from django.views.decorators.http import require_POST

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__, forms
from ledger.api.helpers import get_character_or_none
from ledger.helpers.core import add_info_to_context
from ledger.models.planetary import CharacterPlanetDetails

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("ledger.basic_access")
def planetary_ledger_index(request):
    """Character Ledger Index View"""
    context = {}
    context = add_info_to_context(request, context)
    return redirect(
        "ledger:planetary_ledger", request.user.profile.main_character.character_id
    )


@login_required
@permission_required(["ledger.basic_access"])
def planetary_ledger(request, character_id=None):
    if character_id is None:
        character_id = request.user.profile.main_character.character_id

    context = {
        "title": "Planetary Ledger",
        "character_id": character_id,
        "forms": {
            "confirm": forms.ConfirmForm(),
        },
    }
    context = add_info_to_context(request, context)
    return render(
        request, "ledger/charledger/planetary/planetary_ledger.html", context=context
    )


@login_required
@permission_required("ledger.basic_access")
def planetary_overview(request):
    """
    Planetary Overview
    """

    context = {
        "title": "Planetary Overview",
    }
    context = add_info_to_context(request, context)

    return render(
        request,
        "ledger/charledger/planetary/admin/planetary_overview.html",
        context=context,
    )


@login_required
@permission_required("ledger.basic_access")
@require_POST
def switch_alarm(request):
    # Check Permission
    form = forms.ConfirmForm(request.POST)

    if form.is_valid():
        is_all = False
        character_id = int(form.cleaned_data["character_id"])
        planet_id = int(form.cleaned_data["planet_id"])

        if character_id == 0:  # pylint: disable=duplicate-code
            character_id = request.user.profile.main_character.character_id
            is_all = True

        perm, character = get_character_or_none(
            request, character_id
        )  # pylint: disable=duplicate-code

        if not perm:
            msg = _("Permission Denied")
            return JsonResponse(
                {"success": True, "message": msg},
                status=HTTPStatus.FORBIDDEN,
                safe=False,
            )

        if is_all:
            characters = character.alts.values_list("character_id", flat=True)
        else:
            characters = [character_id]

        filters = Q(planet__character__eve_character__character_id__in=characters)
        if planet_id != 0:
            filters &= Q(planet__planet__id=planet_id)
        try:
            planets = CharacterPlanetDetails.objects.filter(filters)

            if planets.exists():
                # Determine the majority state
                on_count = planets.filter(notification=True).count()
                off_count = planets.filter(notification=False).count()
                majority_state = on_count > off_count

                # Set all to the opposite of the majority state
                for p in planets:
                    p.notification = not majority_state
                    p.save()

                msg = _("All alarms successfully switched")
            else:
                raise CharacterPlanetDetails.DoesNotExist

        except CharacterPlanetDetails.DoesNotExist:
            msg = _("Planet/s not found")
            return JsonResponse(
                {"success": True, "message": msg},
                status=HTTPStatus.NOT_FOUND,
                safe=False,
            )

        return JsonResponse(
            {"success": True, "message": msg}, status=HTTPStatus.OK, safe=False
        )

    msg = "Invalid Form"
    return JsonResponse(
        {"success": False, "message": msg}, status=HTTPStatus.BAD_REQUEST, safe=False
    )



# Standard Library
from types import SimpleNamespace
from unittest.mock import patch

# Django
from django.test import TestCase
from django.utils import timezone

# Alliance Auth (External Libs)
from eveuniverse.models import EveMarketPrice, EveSolarSystem, EveType

# AA Ledger
from ledger.models.characteraudit import CharacterMiningLedger
from ledger.tests.testdata.generate_characteraudit import (
    add_characteraudit_character_to_user,
    create_user_from_evecharacter_with_access,
)
from ledger.tests.testdata.generate_miningledger import create_miningledger
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveentity import load_eveentity
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "ledger.models.characteraudit"


class TestCharacterMiningLedgerModel(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()

        cls.user, cls.character_ownership = create_user_from_evecharacter_with_access(
            1001,
        )
        cls.audit = add_characteraudit_character_to_user(
            cls.user, cls.character_ownership.character.character_id
        )
        cls.eve_type = EveType.objects.get(id=17425)
        cls.eve_system = EveSolarSystem.objects.get(id=30004783)

        cls.eve_type2 = EveType.objects.get(id=16268)
        cls.eve_type_price = EveType.objects.get(id=28437)

        cls.miningentry = create_miningledger(
            character=cls.audit,
            id=1,
            date=timezone.now(),
            type=cls.eve_type,
            system=cls.eve_system,
            quantity=100,
        )
        cls.miningentry2 = create_miningledger(
            character=cls.audit,
            id=2,
            date=timezone.now(),
            type=cls.eve_type2,
            system=cls.eve_system,
            quantity=100,
        )
        cls.miningrecord = SimpleNamespace(
            date=timezone.datetime.replace(
                timezone.now(),
                year=2024,
                month=1,
                day=1,
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ),
            type_id=1,
            solar_system_id=1,
        )
        cls.eve_market_price = EveMarketPrice.objects.create(
            eve_type=cls.eve_type_price,
            average_price=100,
        )

    def test_str(self):
        self.assertEqual(str(self.miningentry), f"{self.audit} 1")

    def test_create_primary_key(self):
        # when
        primary_key = CharacterMiningLedger.create_primary_key(
            self.audit.eve_character.character_id, self.miningrecord
        )
        # then
        self.assertEqual(primary_key, "20240101-1-1001-1")

    def test_get_npc_price(self):
        # when
        npc_price = self.miningentry2.get_npc_price()
        # then
        self.assertIsNotNone(npc_price)
        self.assertEqual(npc_price, 100)

    @patch(MODULE_PATH + ".cache.get", return_value=False)
    @patch(MODULE_PATH + ".EveMarketPrice.objects.update_from_esi")
    def test_update_evemarket_price(self, mock_market_price, mock_cache_get):
        # given
        mock_market_price.return_value = 1337
        # when
        updated = self.miningentry.update_evemarket_price()
        # then
        self.assertTrue(mock_market_price.called)
        self.assertEqual(updated, 1337)



# Generated by Django 4.2.11 on 2024-08-30 14:51

# Django
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eveuniverse", "0010_alter_eveindustryactivityduration_eve_type_and_more"),
        ("ledger", "0003_alter_characteraudit_options_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="characteraudit",
            name="last_update_planetary",
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.CreateModel(
            name="CharacterPlanet",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "upgrade_level",
                    models.IntegerField(
                        default=0, help_text="Upgrade level of the planet"
                    ),
                ),
                (
                    "num_pins",
                    models.IntegerField(
                        default=0, help_text="Number of pins on the planet"
                    ),
                ),
                (
                    "last_update",
                    models.DateTimeField(blank=True, default=None, null=True),
                ),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ledger_characterplanet",
                        to="ledger.characteraudit",
                    ),
                ),
                (
                    "planet",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ledger_planet",
                        to="eveuniverse.eveplanet",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="CharacterPlanetDetails",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                ("links", models.JSONField(blank=True, default=None, null=True)),
                ("pins", models.JSONField(blank=True, default=None, null=True)),
                ("routes", models.JSONField(blank=True, default=None, null=True)),
                (
                    "last_update",
                    models.DateTimeField(blank=True, default=None, null=True),
                ),
                (
                    "last_alert",
                    models.DateTimeField(blank=True, default=None, null=True),
                ),
                ("notification", models.BooleanField(default=False)),
                ("notification_sent", models.BooleanField(default=False)),
                (
                    "planet",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ledger_characterplanetdetails",
                        to="ledger.characterplanet",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
                "indexes": [
                    models.Index(
                        fields=["planet"], name="ledger_char_planet__6169b9_idx"
                    )
                ],
            },
        ),
        migrations.AddIndex(
            model_name="characterplanet",
            index=models.Index(
                fields=["character"], name="ledger_char_charact_f3af73_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="characterplanet",
            index=models.Index(
                fields=["planet"], name="ledger_char_planet__58a5b6_idx"
            ),
        ),
    ]



"""Helpers for wallet journals."""

# Standard Library
import enum
from typing import TYPE_CHECKING

# Django
from django.utils.translation import gettext as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__

if TYPE_CHECKING:
    # AA Ledger
    from ledger.helpers.core import LedgerEntity

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


# Unified Journal Reference Type Enum - All ref types in one place
class JournalRefType(enum.Enum):
    """All wallet journal reference types unified."""

    # Original ref types
    PLAYER_TRADING = 1
    MARKET_TRANSACTION = 2
    GM_CASH_TRANSFER = 3
    MISSION_REWARD = 7
    CLONE_ACTIVATION = 8
    INHERITANCE = 9
    PLAYER_DONATION = 10
    CORPORATION_PAYMENT = 11
    DOCKING_FEE = 12
    OFFICE_RENTAL_FEE = 13
    FACTORY_SLOT_RENTAL_FEE = 14
    REPAIR_BILL = 15
    BOUNTY = 16
    BOUNTY_PRIZE = 17
    INSURANCE = 19
    MISSION_EXPIRATION = 20
    MISSION_COMPLETION = 21
    SHARES = 22
    COURIER_MISSION_ESCROW = 23
    MISSION_COST = 24
    AGENT_MISCELLANEOUS = 25
    LP_STORE = 26
    AGENT_LOCATION_SERVICES = 27
    AGENT_DONATION = 28
    AGENT_SECURITY_SERVICES = 29
    AGENT_MISSION_COLLATERAL_PAID = 30
    AGENT_MISSION_COLLATERAL_REFUNDED = 31
    AGENTS_PREWARD = 32
    AGENT_MISSION_REWARD = 33
    AGENT_MISSION_TIME_BONUS_REWARD = 34
    CSPA = 35
    CSPAOFFLINEREFUND = 36
    CORPORATION_ACCOUNT_WITHDRAWAL = 37
    CORPORATION_DIVIDEND_PAYMENT = 38
    CORPORATION_REGISTRATION_FEE = 39
    CORPORATION_LOGO_CHANGE_COST = 40
    RELEASE_OF_IMPOUNDED_PROPERTY = 41
    MARKET_ESCROW = 42
    AGENT_SERVICES_RENDERED = 43
    MARKET_FINE_PAID = 44
    CORPORATION_LIQUIDATION = 45
    BROKERS_FEE = 46
    CORPORATION_BULK_PAYMENT = 47
    ALLIANCE_REGISTRATION_FEE = 48
    WAR_FEE = 49
    ALLIANCE_MAINTAINANCE_FEE = 50
    CONTRABAND_FINE = 51
    CLONE_TRANSFER = 52
    ACCELERATION_GATE_FEE = 53
    TRANSACTION_TAX = 54
    JUMP_CLONE_INSTALLATION_FEE = 55
    MANUFACTURING = 56
    RESEARCHING_TECHNOLOGY = 57
    RESEARCHING_TIME_PRODUCTIVITY = 58
    RESEARCHING_MATERIAL_PRODUCTIVITY = 59
    COPYING = 60
    REVERSE_ENGINEERING = 62
    CONTRACT_AUCTION_BID = 63
    CONTRACT_AUCTION_BID_REFUND = 64
    CONTRACT_COLLATERAL = 65
    CONTRACT_REWARD_REFUND = 66
    CONTRACT_AUCTION_SOLD = 67
    CONTRACT_REWARD = 68
    CONTRACT_COLLATERAL_REFUND = 69
    CONTRACT_COLLATERAL_PAYOUT = 70
    CONTRACT_PRICE = 71
    CONTRACT_BROKERS_FEE = 72
    CONTRACT_SALES_TAX = 73
    CONTRACT_DEPOSIT = 74
    CONTRACT_DEPOSIT_SALES_TAX = 75
    CONTRACT_AUCTION_BID_CORP = 77
    CONTRACT_COLLATERAL_DEPOSITED_CORP = 78
    CONTRACT_PRICE_PAYMENT_CORP = 79
    CONTRACT_BROKERS_FEE_CORP = 80
    CONTRACT_DEPOSIT_CORP = 81
    CONTRACT_DEPOSIT_REFUND = 82
    CONTRACT_REWARD_DEPOSITED = 83
    CONTRACT_REWARD_DEPOSITED_CORP = 84
    BOUNTY_PRIZES = 85
    ADVERTISEMENT_LISTING_FEE = 86
    MEDAL_CREATION = 87
    MEDAL_ISSUED = 88
    DNA_MODIFICATION_FEE = 90
    SOVEREIGNITY_BILL = 91
    BOUNTY_PRIZE_CORPORATION_TAX = 92
    AGENT_MISSION_REWARD_CORPORATION_TAX = 93
    AGENT_MISSION_TIME_BONUS_REWARD_CORPORATION_TAX = 94
    UPKEEP_ADJUSTMENT_FEE = 95
    PLANETARY_IMPORT_TAX = 96
    PLANETARY_EXPORT_TAX = 97
    PLANETARY_CONSTRUCTION = 98
    CORPORATE_REWARD_PAYOUT = 99
    BOUNTY_SURCHARGE = 101
    CONTRACT_REVERSAL = 102
    CORPORATE_REWARD_TAX = 103
    STORE_PURCHASE = 106
    STORE_PURCHASE_REFUND = 107
    DATACORE_FEE = 112
    WAR_FEE_SURRENDER = 113
    WAR_ALLY_CONTRACT = 114
    BOUNTY_REIMBURSEMENT = 115
    KILL_RIGHT_FEE = 116
    SECURITY_PROCESSING_FEE = 117
    INDUSTRY_JOB_TAX = 120
    INFRASTRUCTURE_HUB_MAINTENANCE = 122
    ASSET_SAFETY_RECOVERY_TAX = 123
    OPPORTUNITY_REWARD = 124
    PROJECT_DISCOVERY_REWARD = 125
    PROJECT_DISCOVERY_TAX = 126
    REPROCESSING_TAX = 127
    JUMP_CLONE_ACTIVATION_FEE = 128
    OPERATION_BONUS = 129
    RESOURCE_WARS_REWARD = 131
    DUEL_WAGER_ESCROW = 132
    DUEL_WAGER_PAYMENT = 133
    DUEL_WAGER_REFUND = 134
    REACTION = 135

    # V3 additions
    STRUCTURE_GATE_JUMP = 140

    # V4 additions
    EXTERNAL_TRADE_FREEZE = 136
    EXTERNAL_TRADE_THAW = 137
    EXTERNAL_TRADE_DELIVERY = 138
    SEASON_CHALLENGE_REWARD = 139
    SKILL_PURCHASE = 141
    ITEM_TRADER_PAYMENT = 142
    FLUX_TICKET_SALE = 143
    FLUX_PAYOUT = 144
    FLUX_TAX = 145
    FLUX_TICKET_REPAYMENT = 146
    REDEEMED_ISK_TOKEN = 147
    DAILY_CHALLENGE_REWARD = 148
    MARKET_PROVIDER_TAX = 149
    ESS_ESCROW_TRANSFER = 155
    MILESTONE_REWARD_PAYMENT = 156
    UNDER_CONSTRUCTION = 166
    ALLIGNMENT_BASED_GATE_TOLL = 168
    PROJECT_PAYOUTS = 170
    INSURGENCY_CORRUPTION_CONTRIBUTION_REWARD = 172
    INSURGENCY_SUPPRESSION_CONTRIBUTION_REWARD = 173
    DAILY_GOAL_PAYOUTS = 174
    DAILY_GOAL_PAYOUTS_TAX = 175
    COSMETIC_MARKET_COMPONENT_ITEM_PURCHASE = 178
    COSMETIC_MARKET_SKIN_SALE_BROKER_FEE = 179
    COSMETIC_MARKET_SKIN_PURCHASE = 180
    COSMETIC_MARKET_SKIN_SALE = 181
    COSMETIC_MARKET_SKIN_SALE_TAX = 182
    COSMETIC_MARKET_SKIN_TRANSACTION = 183
    SKYHOOK_CLAIM_FEE = 184
    AIR_CAREER_PROGRAM_REWARD = 185
    FREELANCE_JOBS_DURATION_FEE = 186
    FREELANCE_JOBS_BROADCASTING_FEE = 187
    FREELANCE_JOBS_REWARD_ESCROW = 188
    FREELANCE_JOBS_REWARD = 189
    FREELANCE_JOBS_ESCROW_REFUND = 190
    FREELANCE_JOBS_REWARD_CORPORATION_TAX = 191
    GM_PLEX_FEE_REFUND = 192


class RefTypeManager:
    """Categories for wallet journal reference types."""

    # Translations for reference types
    _("Corporation Contract")
    _("Corporation Withdrawal")
    _("Mission Reward")
    _("Market")
    _("Daily Goal Reward")
    _("Structure Rental")

    # Assets/Items
    ASSETS = [
        JournalRefType.REPAIR_BILL.name.lower(),
        JournalRefType.ASSET_SAFETY_RECOVERY_TAX.name.lower(),
        JournalRefType.REPROCESSING_TAX.name.lower(),
        JournalRefType.ITEM_TRADER_PAYMENT.name.lower(),
    ]

    # PVE Income
    BOUNTY_PRIZES = [
        JournalRefType.BOUNTY.name.lower(),
        JournalRefType.BOUNTY_PRIZE.name.lower(),
        JournalRefType.BOUNTY_PRIZES.name.lower(),
        JournalRefType.BOUNTY_REIMBURSEMENT.name.lower(),
        JournalRefType.BOUNTY_SURCHARGE.name.lower(),
    ]

    CONTRACT = [
        JournalRefType.CONTRACT_AUCTION_BID.name.lower(),
        JournalRefType.CONTRACT_AUCTION_BID_REFUND.name.lower(),
        JournalRefType.CONTRACT_COLLATERAL.name.lower(),
        JournalRefType.CONTRACT_REWARD_REFUND.name.lower(),
        JournalRefType.CONTRACT_AUCTION_SOLD.name.lower(),
        JournalRefType.CONTRACT_REWARD.name.lower(),
        JournalRefType.CONTRACT_COLLATERAL_REFUND.name.lower(),
        JournalRefType.CONTRACT_COLLATERAL_PAYOUT.name.lower(),
        JournalRefType.CONTRACT_PRICE.name.lower(),
        JournalRefType.CONTRACT_BROKERS_FEE.name.lower(),
        JournalRefType.CONTRACT_SALES_TAX.name.lower(),
        JournalRefType.CONTRACT_DEPOSIT.name.lower(),
        JournalRefType.CONTRACT_DEPOSIT_SALES_TAX.name.lower(),
        JournalRefType.CONTRACT_AUCTION_BID_CORP.name.lower(),
        JournalRefType.CONTRACT_COLLATERAL_DEPOSITED_CORP.name.lower(),
        JournalRefType.CONTRACT_DEPOSIT_CORP.name.lower(),
        JournalRefType.CONTRACT_DEPOSIT_REFUND.name.lower(),
        JournalRefType.CONTRACT_REWARD_DEPOSITED.name.lower(),
        JournalRefType.CONTRACT_REWARD_DEPOSITED_CORP.name.lower(),
        JournalRefType.CONTRACT_REVERSAL.name.lower(),
    ]

    CORPORATION_ADMINISTRATION = [
        JournalRefType.CORPORATION_DIVIDEND_PAYMENT.name.lower(),
        JournalRefType.CORPORATION_REGISTRATION_FEE.name.lower(),
        JournalRefType.CORPORATION_LOGO_CHANGE_COST.name.lower(),
        JournalRefType.CORPORATION_BULK_PAYMENT.name.lower(),
        JournalRefType.ADVERTISEMENT_LISTING_FEE.name.lower(),
    ]

    CORPORATION_CONTRACT = [
        JournalRefType.CONTRACT_PRICE_PAYMENT_CORP.name.lower(),
    ]

    CORPORATION_DONATION = [
        JournalRefType.CORPORATION_ACCOUNT_WITHDRAWAL.name.lower(),
    ]

    DAILY_GOAL_REWARD = [
        JournalRefType.DAILY_GOAL_PAYOUTS.name.lower(),
        JournalRefType.DAILY_CHALLENGE_REWARD.name.lower(),
        JournalRefType.SEASON_CHALLENGE_REWARD.name.lower(),
        JournalRefType.OPPORTUNITY_REWARD.name.lower(),
        JournalRefType.AIR_CAREER_PROGRAM_REWARD.name.lower(),
    ]

    DONATION = [
        JournalRefType.PLAYER_DONATION.name.lower(),
        JournalRefType.AGENT_DONATION.name.lower(),
    ]

    ESS_TRANSFER = [
        JournalRefType.ESS_ESCROW_TRANSFER.name.lower(),
    ]

    FREELANCE_JOBS = [
        JournalRefType.FREELANCE_JOBS_DURATION_FEE.name.lower(),
        JournalRefType.FREELANCE_JOBS_BROADCASTING_FEE.name.lower(),
        JournalRefType.FREELANCE_JOBS_REWARD_ESCROW.name.lower(),
        JournalRefType.FREELANCE_JOBS_REWARD.name.lower(),
        JournalRefType.FREELANCE_JOBS_ESCROW_REFUND.name.lower(),
        JournalRefType.FREELANCE_JOBS_REWARD_CORPORATION_TAX.name.lower(),
    ]

    INCURSION = [
        JournalRefType.RESOURCE_WARS_REWARD.name.lower(),
        JournalRefType.INSURGENCY_CORRUPTION_CONTRIBUTION_REWARD.name.lower(),
        JournalRefType.INSURGENCY_SUPPRESSION_CONTRIBUTION_REWARD.name.lower(),
        JournalRefType.CORPORATE_REWARD_PAYOUT.name.lower(),
    ]

    INSURANCE = [
        JournalRefType.INSURANCE.name.lower(),
    ]

    LP = [
        JournalRefType.LP_STORE.name.lower(),
    ]

    MARKET = [
        JournalRefType.MARKET_TRANSACTION.name.lower(),
        JournalRefType.MARKET_ESCROW.name.lower(),
        JournalRefType.BROKERS_FEE.name.lower(),
        JournalRefType.TRANSACTION_TAX.name.lower(),
        JournalRefType.MARKET_FINE_PAID.name.lower(),
        JournalRefType.MARKET_PROVIDER_TAX.name.lower(),
    ]

    MILESTONE_REWARD = [
        JournalRefType.MILESTONE_REWARD_PAYMENT.name.lower(),
        JournalRefType.PROJECT_DISCOVERY_REWARD.name.lower(),
        JournalRefType.PROJECT_PAYOUTS.name.lower(),
    ]

    MISSION_REWARD = [
        JournalRefType.MISSION_REWARD.name.lower(),
        JournalRefType.MISSION_COMPLETION.name.lower(),
        JournalRefType.AGENT_MISSION_REWARD.name.lower(),
        JournalRefType.AGENT_MISSION_TIME_BONUS_REWARD.name.lower(),
        JournalRefType.AGENTS_PREWARD.name.lower(),
    ]

    PLANETARY = [
        JournalRefType.PLANETARY_IMPORT_TAX.name.lower(),
        JournalRefType.PLANETARY_EXPORT_TAX.name.lower(),
        JournalRefType.PLANETARY_CONSTRUCTION.name.lower(),
    ]

    PRODUCTION = [
        JournalRefType.MANUFACTURING.name.lower(),
        JournalRefType.RESEARCHING_TECHNOLOGY.name.lower(),
        JournalRefType.RESEARCHING_TIME_PRODUCTIVITY.name.lower(),
        JournalRefType.RESEARCHING_MATERIAL_PRODUCTIVITY.name.lower(),
        JournalRefType.COPYING.name.lower(),
        JournalRefType.REVERSE_ENGINEERING.name.lower(),
        JournalRefType.INDUSTRY_JOB_TAX.name.lower(),
        JournalRefType.REACTION.name.lower(),
    ]

    STRUCTURE_RENTAL = [
        JournalRefType.OFFICE_RENTAL_FEE.name.lower(),
        JournalRefType.FACTORY_SLOT_RENTAL_FEE.name.lower(),
        JournalRefType.SOVEREIGNITY_BILL.name.lower(),
        JournalRefType.INFRASTRUCTURE_HUB_MAINTENANCE.name.lower(),
    ]

    SKILL = [
        JournalRefType.SKILL_PURCHASE.name.lower(),
        JournalRefType.DATACORE_FEE.name.lower(),
    ]

    TRAVELING = [
        JournalRefType.DOCKING_FEE.name.lower(),
        JournalRefType.ACCELERATION_GATE_FEE.name.lower(),
        JournalRefType.JUMP_CLONE_INSTALLATION_FEE.name.lower(),
        JournalRefType.JUMP_CLONE_ACTIVATION_FEE.name.lower(),
        JournalRefType.CLONE_ACTIVATION.name.lower(),
        JournalRefType.CLONE_TRANSFER.name.lower(),
        JournalRefType.STRUCTURE_GATE_JUMP.name.lower(),
        JournalRefType.ALLIGNMENT_BASED_GATE_TOLL.name.lower(),
    ]

    @classmethod
    def get_ref_types_from_category(cls, category: str) -> list[str]:
        """Get all ref types from a specific category."""
        if hasattr(cls, category):
            return getattr(cls, category)
        logger.debug(f"Category '{category}' does not exist in RefTypeCategories.")
        return []

    @classmethod
    def get_all_categories(cls) -> dict[str, list[str]]:
        """Get all categories and their ref types, sorted alphabetically by key in the dict literal. Add NO_CATEGORY for missing JournalRefType."""
        categories = {
            "ASSETS": cls.ASSETS,
            "CONTRACT": cls.CONTRACT,
            "CORPORATION_ADMINISTRATION": cls.CORPORATION_ADMINISTRATION,
            "CORPORATION_CONTRACT": cls.CORPORATION_CONTRACT,
            "CORPORATION_WITHDRAWAL": cls.CORPORATION_DONATION,
            "DAILY_GOAL_REWARD": cls.DAILY_GOAL_REWARD,
            "DONATION": cls.DONATION,
            "FREELANCE_JOBS": cls.FREELANCE_JOBS,
            "INCURSION": cls.INCURSION,
            "INSURANCE": cls.INSURANCE,
            "LP": cls.LP,
            "MARKET": cls.MARKET,
            "MISSION_REWARD": cls.MISSION_REWARD,
            "PLANETARY": cls.PLANETARY,
            "PRODUCTION": cls.PRODUCTION,
            "SKILL": cls.SKILL,
            "STRUCTURE_RENTAL": cls.STRUCTURE_RENTAL,
            "TRAVELING": cls.TRAVELING,
        }

        # Alle zugeordneten ref_types sammeln
        assigned = set()
        for ref_types in categories.values():
            assigned.update(ref_types)

        # Alle JournalRefType-Namen in Kleinbuchstaben
        all_types = {jt.name.lower() for jt in JournalRefType}

        # Exclude Bounty and ESS
        pve = set()
        # Bounty-Prizes
        if hasattr(cls, "BOUNTY_PRIZES"):
            pve.update(cls.BOUNTY_PRIZES)
        # ESS Transfer
        if hasattr(cls, "ESS_TRANSFER"):
            pve.update(cls.ESS_TRANSFER)

        # Nicht zugeordnete Typen bestimmen, aber special auslassen
        not_defined = sorted((all_types - assigned) - pve)
        if not_defined:
            categories["NOT_DEFINED_CATEGORY"] = not_defined

        return categories

    @classmethod
    def all_ref_types(cls) -> list[str]:
        """Get all ref types from all categories."""
        all_ref_types = cls.get_all_categories()
        ref_types_items = []
        for __, ref_types in all_ref_types.items():
            ref_types_items.extend(ref_types)
        return ref_types_items

    @staticmethod
    def special_cases(row: dict, ids: set[int], account_char_ids: set[int]) -> bool:
        """Handle special cases in Ledger."""
        if isinstance(row, dict) is False:
            logger.debug("Row is not a dictionary, skipping special case checks.")
            return False

        # Skip Market Transactions from buyer between the corporation and its members (only count transactions from creator)
        if row["ref_type"] == "market_transaction" and row["first_party_id"] in ids:
            return True

        # Skip Contract if Contract Creator is Registered as a Member of the Corporation (only count the contract creator)
        if (
            row["ref_type"] == "contract_price_payment_corp"
            and row["first_party_id"] in account_char_ids
            and row["second_party_id"] in ids
        ):
            return True
        return False

    @staticmethod
    def special_cases_details(
        value: list,
        entity: "LedgerEntity",
        kwargs: dict[str, int],
        journal_type: str,
        char_ids: set[int] = None,
    ) -> bool:
        """Handle special cases in Ledger for Details View."""
        # Skip Contract if Contract Creator is Registered as a Member of the Corporation (only count the contract creator)
        if (
            "contract_price_payment_corp" in value
            and entity.type == "character"
            and journal_type == "corporation"
        ):  # Only Count Contract Creator
            kwargs["first_party"] = entity.entity_id
            return kwargs

        # Skip Player Donation if it is to own alts
        if (
            "player_donation" in value
            and entity.is_eve_character
            and journal_type == "character"
        ):
            if char_ids is None:
                return kwargs
            kwargs["exclude"] = char_ids
        return kwargs



"""PvE Views"""

# Standard Library
import json
from http import HTTPStatus

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.core.handlers.wsgi import WSGIRequest
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.utils.translation import gettext as _
from django.views.decorators.http import require_POST

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.api.helpers import get_character_or_none
from ledger.helpers.character import CharacterData
from ledger.helpers.core import add_info_to_context
from ledger.models.characteraudit import CharacterAudit, CharacterWalletJournalEntry

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("ledger.basic_access")
# pylint: disable=too-many-positional-arguments
def character_ledger(
    request: WSGIRequest,
    character_id: int,
    year=None,
    month=None,
    day=None,
    section=None,
):
    """
    Character Ledger
    """
    perms, character = get_character_or_none(request, character_id)

    if not perms:
        msg = _("Permission Denied")
        messages.error(request, msg)
        context = {
            "error": msg,
            "character_id": character_id,
            "disabled": True,
        }
        context = add_info_to_context(request, context)
        return render(
            request, "ledger/charledger/character_ledger.html", context=context
        )

    character_data = CharacterData(
        request=request,
        character=character,
        year=year,
        month=month,
        day=day,
        section=section,
    )

    # Create the ledger data
    ledger = character_data.generate_ledger_data()

    context = {
        "title": f"Character Ledger - {character.eve_character.character_name}",
        "character_id": character_id,
        "billboard": json.dumps(character_data.billboard.dict.asdict()),
        "ledger": ledger,
        "years": CharacterWalletJournalEntry.objects.filter(
            character__in=character_data.characters
        )
        .exclude(date__year__isnull=True)
        .values_list("date__year", flat=True)
        .order_by("-date__year")
        .distinct(),
        "totals": character_data.calculate_totals(ledger),
        "view": character_data.create_view_data(
            viewname="character_details",
            character_id=character_id,
            section="summary",
        ),
        "is_old_ess": character_data.is_old_ess,
    }
    # Add additional information to the context
    context = add_info_to_context(request, context)

    return render(request, "ledger/charledger/character_ledger.html", context=context)


@login_required
@permission_required("ledger.basic_access")
# pylint: disable=too-many-positional-arguments
def character_details(
    request, character_id, year=None, month=None, day=None, section=None
):
    """
    Character Details
    """
    perms, character = get_character_or_none(request, character_id)

    # pylint: disable=duplicate-code
    if perms is False:
        msg = _("Permission Denied")
        return render(
            request,
            "ledger/partials/information/view_character_content.html",
            {
                "error": msg,
                "character_id": character_id,
            },
        )
    # pylint: disable=duplicate-code
    if perms is None:
        msg = _("Corporation not found")
        return render(
            request,
            "ledger/partials/information/view_character_content.html",
            {
                "error": msg,
                "character_id": character_id,
            },
        )

    character_data = CharacterData(request, character, year, month, day, section)

    journal, mining = character_data.filter_character_journal(character)
    amounts = character_data._create_character_details(journal, mining)
    details = character_data._add_average_details(request, amounts, day)

    context = {
        "title": f"Character Details - {character.eve_character.character_name}",
        "type": "character",
        "character": details,
        "information": f"Character Details - {character_data.get_details_title}",
    }
    context = add_info_to_context(request, context)

    return render(
        request,
        "ledger/partials/information/view_character_content.html",
        context=context,
    )


@login_required
@permission_required("ledger.basic_access")
def character_overview(request):
    """
    Character Overview
    """

    context = {
        "title": "Character Overview",
    }
    context = add_info_to_context(request, context)
    return render(
        request, "ledger/charledger/admin/character_overview.html", context=context
    )


@login_required
@permission_required("ledger.basic_access")
def character_administration(request, character_id=None):
    """
    Character Administration
    """
    if character_id is None:
        character_id = request.user.profile.main_character.character_id

    perms, character = get_character_or_none(request, character_id)

    if not perms:
        msg = _("Permission Denied")
        messages.error(request, msg)
        return redirect("ledger:index")

    linked_characters_ids = character.alts.values_list("character_id", flat=True)

    characters = CharacterAudit.objects.filter(
        eve_character__character_id__in=linked_characters_ids
    ).order_by("eve_character__character_name")
    characters_ids = characters.values_list("eve_character__character_id", flat=True)

    missing_characters = (
        EveCharacter.objects.filter(character_id__in=linked_characters_ids)
        .exclude(character_id__in=characters_ids)
        .order_by("character_name")
    )

    context = {
        "character_id": character_id,
        "title": "Character Administration",
        "characters": characters,
        "missing_characters": missing_characters,
    }
    context = add_info_to_context(request, context)
    return render(
        request,
        "ledger/charledger/admin/character_administration.html",
        context=context,
    )


@login_required
@permission_required("ledger.basic_access")
@require_POST
def character_delete(request, character_id):
    """
    Character Delete
    """
    perms = get_character_or_none(request, character_id)[0]
    if not perms:
        msg = _("Permission Denied")
        return JsonResponse(
            {"success": False, "message": msg}, status=HTTPStatus.FORBIDDEN, safe=False
        )

    try:
        audit = CharacterAudit.objects.get(eve_character__character_id=character_id)
    except CharacterAudit.DoesNotExist:
        msg = _("Character not found")
        return JsonResponse(
            {"success": False, "message": msg}, status=HTTPStatus.NOT_FOUND, safe=False
        )

    audit.delete()

    msg = _(f"{audit.eve_character.character_name} successfully deleted")
    return JsonResponse(
        {"success": True, "message": msg}, status=HTTPStatus.OK, safe=False
    )



# Third Party
from charlink.app_imports.utils import AppImport, LoginImport

# Django
from django.contrib.auth.models import Permission, User
from django.db.models import Exists, OuterRef
from django.utils.translation import gettext_lazy as trans

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo

# Alliance Auth (External Libs)
from app_utils.allianceauth import users_with_permission

# AA Ledger
from ledger.app_settings import LEDGER_APP_NAME
from ledger.models.characteraudit import CharacterAudit
from ledger.models.corporationaudit import CorporationAudit
from ledger.tasks import update_character, update_corporation

_corp_perms = [
    "ledger.manage_access",
]


# pylint: disable=unused-argument, duplicate-code
def _add_character_charaudit(request, token):
    character = CharacterAudit.objects.update_or_create(
        eve_character=EveCharacter.objects.get_character_by_id(token.character_id),
        defaults={
            "character_name": token.character_name,
        },
    )[0]
    update_character.apply_async(
        args=[character.pk], kwargs={"force_refresh": True}, priority=6
    )


# pylint: disable=unused-argument, duplicate-code
def _add_character_corp(request, token):
    char = EveCharacter.objects.get_character_by_id(token.character_id)
    eve_corp, _ = EveCorporationInfo.objects.get_or_create(
        corporation_id=char.corporation_id,
        defaults={
            "member_count": 0,
            "corporation_ticker": char.corporation_ticker,
            "corporation_name": char.corporation_name,
        },
    )
    corp = CorporationAudit.objects.update_or_create(
        corporation=eve_corp,
        defaults={
            "corporation_name": eve_corp.corporation_name,
        },
    )[0]
    update_corporation.apply_async(
        args=[corp.pk], kwargs={"force_refresh": True}, priority=6
    )


def _check_perms_corp(user: User):
    return any(user.has_perm(perm) for perm in _corp_perms)


def _is_character_added_charaudit(character: EveCharacter):
    return CharacterAudit.objects.filter(character=character).exists()


def _is_character_added_corp(character: EveCharacter):
    return CorporationAudit.objects.filter(
        corporation__corporation_id=character.corporation_id
    ).exists()


def _users_with_perms_charaudit():
    return users_with_permission(
        Permission.objects.get(
            content_type__app_label="ledger", codename="basic_access"
        )
    )


def _users_with_perms_corp():
    users = users_with_permission(
        Permission.objects.get(
            content_type__app_label=_corp_perms[0].split(".", maxsplit=1)[0],
            codename=_corp_perms[0].split(".", maxsplit=1)[1],
        )
    )
    for perm_str in _corp_perms[1:]:
        users |= users_with_permission(
            Permission.objects.get(
                content_type__app_label=perm_str.split(".", maxsplit=1)[0],
                codename=perm_str.split(".", maxsplit=1)[1],
            )
        )

    return users


app_import = AppImport(
    "ledger",
    [
        LoginImport(
            app_label="ledger",
            unique_id="default",
            field_label=LEDGER_APP_NAME + " - " + trans("Character") + " Login",
            add_character=_add_character_charaudit,
            scopes=CharacterAudit.get_esi_scopes(),
            check_permissions=lambda user: user.has_perm("ledger.basic_access"),
            is_character_added=_is_character_added_charaudit,
            is_character_added_annotation=Exists(
                CharacterAudit.objects.filter(eve_character_id=OuterRef("pk"))
            ),
            get_users_with_perms=_users_with_perms_charaudit,
        ),
        LoginImport(
            app_label="ledger",
            unique_id="corpaudit",
            field_label=LEDGER_APP_NAME + " - " + trans("Corporation") + " Login",
            add_character=_add_character_corp,
            scopes=CorporationAudit.get_esi_scopes(),
            check_permissions=_check_perms_corp,
            is_character_added=_is_character_added_corp,
            is_character_added_annotation=Exists(
                CorporationAudit.objects.filter(
                    corporation__corporation_id=OuterRef("corporation_id")
                )
            ),
            get_users_with_perms=_users_with_perms_corp,
        ),
    ],
)



# Third Party
from ninja import NinjaAPI
from ninja.security import django_auth

# Django
from django.conf import settings

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.api import ledger

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


api = NinjaAPI(
    title="Geuthur API",
    version="0.5.0",
    urls_namespace="ledger:api",
    auth=django_auth,
    openapi_url=settings.DEBUG and "/openapi.json" or "",
)

# Add the ledger endpoints
ledger.setup(api)



# Generated by Django 4.2.23 on 2025-07-22 04:34

# Django
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        (
            "ledger",
            "0014_alter_corporationaudit_options_alter_general_options_and_more",
        ),
    ]

    operations = [
        migrations.RenameField(
            model_name="characteraudit",
            old_name="character",
            new_name="eve_character",
        ),
    ]



# Django
from django.test import TestCase, modify_settings

# AA Ledger
# AA Skillfarm
from ledger.helpers.discord import (
    allianceauth_discordbot_installed,
    discordnotify_installed,
)


class TestModulesInstalled(TestCase):
    @modify_settings(INSTALLED_APPS={"remove": "aadiscordbot"})
    def test_allianceauth_discordbot_installed_should_return_false(self):
        self.assertFalse(allianceauth_discordbot_installed())

    @modify_settings(INSTALLED_APPS={"append": "aadiscordbot"})
    def test_allianceauth_discordbot_installed_should_return_true(self):
        self.assertTrue(allianceauth_discordbot_installed())

    @modify_settings(INSTALLED_APPS={"remove": "discordnotify"})
    def test_aa_discordnotify_installed_should_return_false(self):
        self.assertFalse(discordnotify_installed())

    @modify_settings(INSTALLED_APPS={"append": "discordnotify"})
    def test_aa_discordnotify_installed_should_return_true(self):
        self.assertTrue(discordnotify_installed())



# Standard Library
from collections import defaultdict
from dataclasses import asdict, dataclass
from decimal import Decimal
from typing import Any

# Django
from django.db.models import QuerySet, TextChoices
from django.db.models.functions import TruncDay, TruncHour, TruncMonth
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@dataclass
class ChartData:
    title: str
    categories: list[str]
    series: list[dict[str, Any]]

    def serialize_decimals(self, obj):
        if isinstance(obj, dict):
            return {k: self.serialize_decimals(v) for k, v in obj.items()}
        if isinstance(obj, list):
            return [self.serialize_decimals(i) for i in obj]
        if isinstance(obj, Decimal):
            return float(obj)
        return obj

    def asdict(self) -> dict:
        """Return this object as dict."""
        serialized_data = self.serialize_decimals(asdict(self))
        return serialized_data


class BillboardSystem:
    """BillboardSystem class to process billboard data."""

    class Categories(TextChoices):
        """Translation Helper"""

        BOUNTY = "Bounty", _("Bounty")
        ESS = "ESS", _("ESS")
        MINING = "Mining", _("Mining")
        MISCELLANEOUS = "Miscellaneous", _("Miscellaneous")
        COSTS = "Costs", _("Costs")
        UNKNOWN = "Unknown", _("Unknown")

    @dataclass
    class BillboardDict:
        """BillboardDict class to store the billboard data."""

        charts: ChartData = None
        rattingbar: ChartData = None

        def asdict(self) -> dict:
            """Return this object as dict."""
            return {
                "charts": self.charts.asdict() if self.charts else None,
                "rattingbar": self.rattingbar.asdict() if self.rattingbar else None,
            }

    def __init__(
        self,
        view: str = "month",
    ):
        self.view = view
        self.dict = self.BillboardDict()
        self.results = defaultdict(lambda: defaultdict(Decimal))

    def _get_formatted_date(self, date, view):
        if view == "year":
            return date.strftime("%Y-%m")
        if view == "month":
            return date.strftime("%Y-%m-%d")
        if self.view == "day":
            return date.strftime("%Y-%m-%d %H:%M")
        raise ValueError("Invalid view type. Use 'day', 'month', or 'year'.")

    def _create_chart_dict(self):
        """Create the Charts dict if it doesn't exist"""
        if self.dict.charts is None:
            self.dict.charts = ChartData(
                title="Billboard",
                categories=[],
                series=[],
            )

    def change_view(self, view: str):
        """Change the view of the billboard"""
        if view not in ["day", "month", "year"]:
            raise ValueError("Invalid view type. Use 'day', 'month', or 'year'.")
        self.view = view

    def chord_add_char_data_from_dict(self, data: dict):
        """Add character data to chord from dict"""
        self._create_chart_dict()

        data_points = [
            {
                "from": f"{data['main_name']}",
                "to": "Wallet",
                "value": abs(data["total_amount"]),
            },
            {
                "from": f"{data['main_name']}",
                "to": "Wallet",
                "value": abs(data["total_amount_ess"]),
            },
            {
                "from": f"{data['main_name']}",
                "to": "Wallet",
                "value": abs(data["total_amount_mining"]),
            },
            {
                "from": f"{data['main_name']}",
                "to": "Wallet",
                "value": abs(data["total_amount_others"]),
            },
            {
                "from": f"{data['main_name']}",
                "to": "Costs",
                "value": abs(data["total_amount_costs"]),
            },
        ]

        for point in data_points:
            if point["value"] != 0:
                self.dict.charts.series.append(point)

    def chord_add_data(self, chord_from: str, chord_to: str, value: int):
        """Add Simple Chord data"""
        self._create_chart_dict()

        if value == 0:
            return

        data = {
            "from": chord_from,
            "to": chord_to,
            "value": value,
        }
        self.dict.charts.series.append(data)

    def chord_handle_overflow(self):
        """Order and handle overflow data for the billboard, for each 'to' category and sort it by from and to."""
        if self.dict.charts is None:
            return

        # Group by 'to' category
        grouped = defaultdict(list)
        for entry in self.dict.charts.series:
            grouped[entry["to"]].append(entry)

        new_series = []
        for to_category, entries in grouped.items():
            # Sort each group by value descending
            sorted_entries = sorted(entries, key=lambda x: x["value"], reverse=True)
            if len(sorted_entries) > 25:
                # Keep top 30, sum the rest as 'Others'
                top_entries = sorted_entries[:25]
                others_value = sum(e["value"] for e in sorted_entries[25:])
                top_entries.append(
                    {
                        "from": "Others",
                        "to": to_category,
                        "value": others_value,
                    }
                )
                new_series.extend(top_entries)
            else:
                new_series.extend(sorted_entries)

        # Sort the final series by value descending for display
        self.dict.charts.series = sorted(
            new_series, key=lambda x: x["value"], reverse=True
        )
        self.sort_chord_data()

    def sort_chord_data(self):
        """Sort the chord data by 'from' and 'to' categories."""
        if self.dict.charts is None:
            return

        self.dict.charts.series.sort(key=lambda x: (x["from"], x["to"]))

    def create_timeline(self, journal: QuerySet):
        """Create the timeline data for the billboard"""
        qs = journal

        if self.view == "year":
            qs = qs.annotate(period=TruncMonth("date"))
        elif self.view == "month":
            qs = qs.annotate(period=TruncDay("date"))
        elif self.view == "day":
            qs = qs.annotate(period=TruncHour("date"))
        else:
            raise ValueError("Invalid view type. Use 'day', 'month', or 'year'.")

        qs = qs.values("period").order_by("period")
        return qs

    def create_or_update_results(
        self,
        qs: QuerySet[dict],
        is_old_ess: bool = False,
    ):
        """Create or update the results for the billboard"""
        for entry in qs:
            date = entry["period"]
            bounty = entry.get("bounty_income", 0)
            ess = entry.get("ess_income", 0)
            miscellaneous = entry.get("miscellaneous", 0)

            self.results[date][self.Categories.BOUNTY] += bounty
            self.results[date][self.Categories.ESS] += (
                ess if not is_old_ess else bounty * Decimal("0.667")
            )
            self.results[date][self.Categories.MISCELLANEOUS] += miscellaneous

        return self.results

    def add_category(self, qs: QuerySet[dict], category: str):
        """Add category data to the results

        **the annotation must be have _income as ending
        """
        for entry in qs:
            date = entry["period"]
            category_value = entry.get(f"{category}_income", 0)
            try:
                self.results[date][self.Categories[category.upper()]] += category_value
            except KeyError:
                self.results[date][self.Categories.UNKNOWN] += category_value

    def generate_xy_series(self):
        """Create the ratting bar amounts and categories for the billboard"""
        series = []
        category_set = set()

        for date, values in self.results.items():
            # Remove categories with value 0
            filtered_values = {str(k): v for k, v in values.items() if v != 0}
            if not filtered_values:
                continue  # Skip if all categories are 0

            series.append(
                {
                    "date": self._get_formatted_date(date, self.view),
                    **{k: int(v) for k, v in filtered_values.items()},
                }
            )
            category_set.update(filtered_values.keys())

        # Create categories
        categories = []
        for cat in sorted(category_set):
            try:
                label = self.Categories(cat).label
            except Exception:  # pylint: disable=broad-except
                label = cat
            categories.append({"name": cat, "label": str(label)})

        if not series:
            return [], []

        # Sort Series by Date
        series.sort(key=lambda x: x["date"])

        return series, categories

    def create_xy_chart(self, title, categories, series):
        """Create the XY chart for the billboard"""
        self.dict.rattingbar = ChartData(
            title=title,
            categories=categories,
            series=series,
        )



# Standard Library
from typing import TYPE_CHECKING

# Django
from django.db import models, transaction
from django.db.models import DecimalField, Q, Sum, Value
from django.db.models.functions import Coalesce

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger
from esi.errors import TokenError

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.decorators import log_timing
from ledger.helpers.ref_type import RefTypeManager
from ledger.providers import esi

if TYPE_CHECKING:
    # Alliance Auth
    from esi.stubs import CharactersCharacterIdWalletJournalGetItem

    # AA Ledger
    from ledger.models.characteraudit import (
        CharacterAudit,
    )
    from ledger.models.general import UpdateSectionResult


logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CharWalletIncomeFilter(models.QuerySet):
    # PvE - Income
    def annotate_bounty_income(self) -> models.QuerySet:
        return self.annotate(
            bounty_income=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.BOUNTY_PRIZES, amount__gt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_ess_income(self) -> models.QuerySet:
        return self.annotate(
            ess_income=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.ESS_TRANSFER, amount__gt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_mission_income(self) -> models.QuerySet:
        return self.annotate(
            mission_income=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.MISSION_REWARD, amount__gt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_incursion_income(self) -> models.QuerySet:
        return self.annotate(
            incursion_income=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.INCURSION, amount__gt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    # Trading - Income
    def annotate_market_income(self) -> models.QuerySet:
        return self.annotate(
            market_income=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.MARKET, amount__gt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_contract_income(self) -> models.QuerySet:
        contract_types = list(RefTypeManager.CONTRACT) + list(
            RefTypeManager.CORPORATION_CONTRACT
        )
        return self.annotate(
            contract_income=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=contract_types, amount__gt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_donation_income(self, exclude: list = None) -> models.QuerySet:
        if exclude is None:
            exclude = []

        return self.annotate(
            donation_income=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.DONATION, amount__gt=0)
                    & ~Q(first_party_id__in=exclude),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_insurance_income(self) -> models.QuerySet:
        return self.annotate(
            insurance_income=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.INSURANCE, amount__gt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_milestone_income(self) -> models.QuerySet:
        return self.annotate(
            milestone_income=Coalesce(
                Sum(
                    "amount",
                    filter=Q(
                        ref_type__in=RefTypeManager.MILESTONE_REWARD, amount__gt=0
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_daily_goal_income(self) -> models.QuerySet:
        """Annotate daily goal income."""
        return self.annotate(
            daily_goal_income=Coalesce(
                Sum(
                    "amount",
                    filter=Q(
                        ref_type__in=RefTypeManager.DAILY_GOAL_REWARD, amount__gt=0
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )


class CharWalletOutSideFilter(CharWalletIncomeFilter):
    def annotate_miscellaneous(self) -> models.QuerySet:
        return self.annotate(
            miscellaneous=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.all_ref_types(), amount__gt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_miscellaneous_exclude_donations(self, exclude=None) -> models.QuerySet:
        """Allow excluding donations by first_party_id from the miscellaneous sum."""
        # Normalize single int to list for compatibility with Q lookups
        if isinstance(exclude, int):
            exclude = [exclude]

        # Ensure we have a mutable list (remove may not exist on other iterables)
        all_types = RefTypeManager.all_ref_types()
        donation_types = RefTypeManager.DONATION

        if exclude is not None:
            try:
                all_types.remove("player_donation")
            except ValueError:
                pass

        # Base filter: all reference types and only positive amounts (income)
        filter_q = Q(ref_type__in=all_types, amount__gt=0)

        if exclude:
            filter_q = filter_q & ~(
                Q(ref_type__in=donation_types) & Q(first_party_id__in=exclude)
            )

        return self.annotate(
            miscellaneous=Coalesce(
                Sum(
                    "amount",
                    filter=filter_q,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_costs(self) -> models.QuerySet:
        return self.annotate(
            costs=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.all_ref_types(), amount__lt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )


class CharWalletCostQueryFilter(CharWalletOutSideFilter):
    # Costs
    def annotate_contract_cost(self) -> models.QuerySet:
        return self.annotate(
            contract_cost=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.CONTRACT, amount__lt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_market_cost(self) -> models.QuerySet:
        return self.annotate(
            market_cost=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.MARKET, amount__lt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_asset_cost(self) -> models.QuerySet:
        return self.annotate(
            asset_cost=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.ASSETS, amount__lt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_traveling_cost(self) -> models.QuerySet:
        return self.annotate(
            traveling_cost=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.TRAVELING, amount__lt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_production_cost(self) -> models.QuerySet:
        return self.annotate(
            production_cost=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.PRODUCTION, amount__lt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_skill_cost(self) -> models.QuerySet:
        return self.annotate(
            skill_cost=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.SKILL, amount__lt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_insurance_cost(self) -> models.QuerySet:
        return self.annotate(
            insurance_cost=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.INSURANCE, amount__lt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_planetary_cost(self) -> models.QuerySet:
        return self.annotate(
            planetary_cost=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.PLANETARY, amount__lt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_lp_cost(self) -> models.QuerySet:
        return self.annotate(
            lp_cost=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.LP, amount__lt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )


# pylint: disable=used-before-assignment
class CharWalletQuerySet(CharWalletCostQueryFilter):
    @log_timing(logger)
    def update_or_create_esi(
        self, character: "CharacterAudit", force_refresh: bool = False
    ) -> "UpdateSectionResult":
        """Update or Create a wallet journal entry from ESI data."""
        return character.update_section_if_changed(
            section=character.UpdateSection.WALLET_JOURNAL,
            fetch_func=self._fetch_esi_data,
            force_refresh=force_refresh,
        )

    def _fetch_esi_data(self, audit: "CharacterAudit", force_refresh: bool) -> None:
        """Fetch wallet journal entries from ESI data."""
        req_scopes = ["esi-wallet.read_character_wallet.v1"]
        token = audit.get_token(scopes=req_scopes)

        operation = esi.client.Wallet.GetCharactersCharacterIdWalletJournal(
            character_id=audit.eve_character.character_id,
            token=token,
        )

        journal_items = operation.results(
            force_refresh=force_refresh,
        )

        self._update_or_create_objs(character=audit, objs=journal_items)

    @transaction.atomic()
    def _update_or_create_objs(
        self,
        character: "CharacterAudit",
        objs: list["CharactersCharacterIdWalletJournalGetItem"],
    ) -> None:
        """Update or Create wallet journal entries from objs data."""
        # pylint: disable=import-outside-toplevel
        # AA Ledger
        from ledger.models.general import EveEntity

        _current_journal = self.filter(character=character).values_list(
            "entry_id", flat=True
        )
        _current_eve_ids = list(
            EveEntity.objects.all().values_list("eve_id", flat=True)
        )

        _new_names = []

        items = []
        for item in objs:
            if item.id not in _current_journal:
                if item.second_party_id not in _current_eve_ids:
                    _new_names.append(item.second_party_id)
                    _current_eve_ids.append(item.second_party_id)
                if item.first_party_id not in _current_eve_ids:
                    _new_names.append(item.first_party_id)
                    _current_eve_ids.append(item.first_party_id)

                asset_item = self.model(
                    character=character,
                    amount=item.amount,
                    balance=item.balance,
                    context_id=item.context_id,
                    context_id_type=item.context_id_type,
                    date=item.date,
                    description=item.description,
                    first_party_id=item.first_party_id,
                    entry_id=item.id,
                    reason=item.reason,
                    ref_type=item.ref_type,
                    second_party_id=item.second_party_id,
                    tax=item.tax,
                    tax_receiver_id=item.tax_receiver_id,
                )
                items.append(asset_item)

        created_names = EveEntity.objects.create_bulk_from_esi(_new_names)

        if created_names:
            self.bulk_create(items)
        else:
            raise TokenError("ESI Fail")

    def aggregate_bounty(self) -> dict:
        """Aggregate bounty income."""
        return self.filter(ref_type__in=RefTypeManager.BOUNTY_PRIZES).aggregate(
            total_bounty=Coalesce(Sum("amount"), Value(0), output_field=DecimalField())
        )["total_bounty"]

    def aggregate_ess(self) -> dict:
        """Aggregate ESS income."""
        return self.filter(ref_type__in=RefTypeManager.ESS_TRANSFER).aggregate(
            total_ess=Coalesce(Sum("amount"), Value(0), output_field=DecimalField())
        )["total_ess"]

    def aggregate_costs(self, first_party=None, second_party=None) -> dict:
        """Aggregate costs. first_party wird für donation exkludiert."""
        qs = self
        cost_types = RefTypeManager.all_ref_types()
        donation_types = RefTypeManager.DONATION

        if first_party is not None:
            if isinstance(first_party, int):
                first_party = [first_party]
            qs = qs.filter(first_party__in=first_party)

        if second_party is not None:
            if isinstance(second_party, int):
                second_party = [second_party]
            # Exclude: alle donation mit first_party_id in Liste, Rest wie gehabt
            qs = qs.exclude(
                Q(ref_type__in=donation_types, first_party_id__in=second_party)
            )
            qs = qs.filter(Q(ref_type__in=cost_types))
        else:
            qs = qs.filter(Q(ref_type__in=cost_types))

        qs = qs.filter(amount__lt=0)
        return qs.aggregate(
            total_costs=Coalesce(Sum("amount"), Value(0), output_field=DecimalField())
        )["total_costs"]

    def aggregate_miscellaneous(self, first_party=None, second_party=None) -> dict:
        """Aggregate miscellaneous income. first_party wird nur für donation angewandt."""
        qs = self

        misc_types = RefTypeManager.all_ref_types()
        donation_types = RefTypeManager.DONATION

        if first_party is not None:
            if isinstance(first_party, int):
                first_party = [first_party]
            # Exclude: alle donation mit first_party_id in Liste, Rest wie gehabt
            qs = qs.exclude(
                Q(ref_type__in=donation_types, first_party_id__in=first_party)
            )
            qs = qs.filter(Q(ref_type__in=misc_types))
        else:
            qs = qs.filter(Q(ref_type__in=misc_types))

        if second_party is not None:
            if isinstance(second_party, int):
                second_party = [second_party]
            qs = qs.filter(second_party__in=second_party)

        qs = qs.filter(amount__gt=0)
        return qs.aggregate(
            total_misc=Coalesce(Sum("amount"), Value(0), output_field=DecimalField())
        )["total_misc"]

    # pylint: disable=too-many-positional-arguments
    def aggregate_ref_type(
        self,
        ref_type: list,
        first_party=None,
        second_party=None,
        exclude=None,
        income: bool = False,
    ) -> dict:
        """Aggregate income by ref_type."""
        qs = self.filter(ref_type__in=ref_type)

        if first_party is not None:
            if isinstance(first_party, int):
                first_party = [first_party]
            qs = qs.filter(first_party__in=first_party)

        if second_party is not None:
            if isinstance(second_party, int):
                second_party = [second_party]
            qs = qs.filter(second_party__in=second_party)

        if exclude is not None:
            if isinstance(exclude, int):
                exclude = [exclude]
            qs = qs.exclude(first_party__in=exclude)

        if income:
            qs = qs.filter(amount__gt=0)
        else:
            qs = qs.filter(amount__lt=0)

        return qs.aggregate(
            total=Coalesce(Sum("amount"), Value(0), output_field=DecimalField())
        )["total"]


class CharWalletManagerBase(models.Manager):
    pass


CharWalletManager = CharWalletManagerBase.from_queryset(CharWalletQuerySet)



# Standard Library
from unittest.mock import patch

# Django
from django.test import TestCase
from django.utils import timezone

# AA Ledger
from ledger.tests.testdata.generate_characteraudit import (
    create_characteraudit_from_evecharacter,
)
from ledger.tests.testdata.generate_planets import (
    _planetary_data,
    create_character_planet,
    create_character_planet_details,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "ledger.models.planetary"


class TestPlanetModel(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.planet_params = {
            "upgrade_level": 5,
            "num_pins": 5,
        }

        cls.audit = create_characteraudit_from_evecharacter(1001)
        cls.planetary = create_character_planet(cls.audit, 4001, **cls.planet_params)

    def test_str(self):
        self.assertEqual(str(self.planetary), "Planet Data: Gneuten - Test Planet I")

    def test_get_esi_scopes(self):
        self.assertEqual(
            self.planetary.get_esi_scopes(), ["esi-planets.manage_planets.v1"]
        )


class TestPlanetaryDetailsModel(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.planet_params = {
            "upgrade_level": 5,
            "num_pins": 5,
        }

        cls.audit = create_characteraudit_from_evecharacter(1001)
        cls.planetary = create_character_planet(cls.audit, 4001, **cls.planet_params)
        cls.planetarydetails = create_character_planet_details(
            cls.planetary, **_planetary_data
        )

    def test_details_str(self):
        self.assertEqual(
            str(self.planetarydetails), "Planet Details Data: Gneuten - Test Planet I"
        )

    def test_count_extractors(self):
        extractors_count = self.planetarydetails.count_extractors()

        self.assertEqual(extractors_count, 2)

    def test_get_planet_install_date(self):
        install_time = self.planetarydetails.get_planet_install_date()

        expected_install_date = timezone.datetime(
            2024, 8, 12, 17, 17, 2, tzinfo=timezone.utc
        )
        self.assertEqual(install_time, expected_install_date)

    def test_get_planet_expiry_date(self):
        expected_expiry_date = timezone.datetime(
            2024, 8, 26, 17, 17, 2, tzinfo=timezone.utc
        )
        self.assertEqual(
            self.planetarydetails.get_planet_expiry_date(), expected_expiry_date
        )

    def test_is_expired(self):
        self.assertEqual(self.planetarydetails.is_expired, True)

    def test_get_types(self):
        self.assertEqual(
            self.planetarydetails.get_types(), [9832, 3645, 2390, 2268, 2309]
        )

    @patch(MODULE_PATH + ".timezone.now")
    def test_is_percent_correct(self, mock_now):
        fixed_date = timezone.make_aware(timezone.datetime(2024, 8, 20, 17, 17, 2))
        mock_now.return_value = fixed_date

        extractor_info = self.planetarydetails.get_extractors_info()
        expected_percent = 57.14
        for _, value in extractor_info.items():
            self.assertIn(
                "progress_percentage", value
            )  # Check if the key exists in the nested dictionary
            self.assertEqual(value["progress_percentage"], expected_percent)

    def test_get_planet_install_date_none(self):
        planetary_details = create_character_planet_details(self.planetary)
        planetary_details.pins = []
        self.assertIsNone(planetary_details.get_planet_install_date())

    def test_get_planet_expiry_date_none(self):
        planetary_details = create_character_planet_details(self.planetary)
        planetary_details.pins = []
        self.assertIsNone(planetary_details.get_planet_expiry_date())

    @patch("django.utils.timezone.now")
    def test_is_expired_false(self, mock_now):
        mock_now.return_value = timezone.datetime(2023, 10, 1, tzinfo=timezone.utc)
        future_date = mock_now.return_value + timezone.timedelta(days=10)

        planetary_details = create_character_planet_details(self.planetary)
        planetary_details.pins = [{"expiry_time": future_date.isoformat()}]
        self.assertFalse(planetary_details.is_expired)

    def test_is_expired_empty(self):
        planetary_details = create_character_planet_details(self.planetary)
        planetary_details.pins = [{"expiry_time": None}]
        self.assertFalse(planetary_details.is_expired)



# Standard Library
import json
from pathlib import Path

# Alliance Auth (External Libs)
from app_utils.esi_testing import EsiClientStub

# AA Ledger
from ledger.tests.testdata.esi_stub_migration import EsiClientStubOpenApi, EsiEndpoint


class CharacterJournalContext:
    """Context for character wallet journal ESI operations."""

    amount: float
    balance: float
    context_id: int
    context_id_type: str
    date: str
    description: str
    first_party_id: int
    id: int
    reason: str
    ref_type: str
    second_party_id: int
    tax: float
    tax_receiver_id: int


def load_test_data():
    file_path = Path(__file__).parent / "esi.json"
    with file_path.open("r", encoding="utf-8") as fp:
        return json.load(fp)


_esi_data = load_test_data()

_endpoints = [
    EsiEndpoint(
        "Character",
        "GetCharactersCharacterIdRoles",
        "character_id",
        needs_token=False,
    ),
    EsiEndpoint(
        "Wallet",
        "GetCharactersCharacterIdWalletJournal",
        "character_id",
        needs_token=False,
    ),
    EsiEndpoint(
        "Wallet",
        "GetCharactersCharacterIdWallet",
        "character_id",
        needs_token=False,
    ),
    EsiEndpoint(
        "Corporation",
        "GetCorporationsCorporationIdDivisions",
        "corporation_id",
        needs_token=False,
    ),
    EsiEndpoint(
        "Wallet",
        "GetCorporationsCorporationIdWallets",
        "corporation_id",
        needs_token=False,
    ),
    EsiEndpoint(
        "Wallet",
        "GetCorporationsCorporationIdWalletsDivisionJournal",
        "corporation_id",
        needs_token=False,
    ),
    EsiEndpoint(
        "Industry",
        "GetCharactersCharacterIdMining",
        "character_id",
        needs_token=False,
    ),
    EsiEndpoint(
        "Planetary_Interaction",
        "GetCharactersCharacterIdPlanets",
        "character_id",
        needs_token=False,
    ),
    EsiEndpoint(
        "Planetary_Interaction",
        "GetCharactersCharacterIdPlanetsPlanetId",
        ("character_id", "planet_id"),
        needs_token=False,
    ),
    EsiEndpoint(
        "Universe",
        "PostUniverseNames",
        "body",
        needs_token=False,
    ),
    EsiEndpoint(
        "Universe",
        "PostUniverseIds",
        "body",
        needs_token=False,
    ),
]

esi_client_stub = EsiClientStub(_esi_data, endpoints=_endpoints)
esi_client_stub_openapi = EsiClientStubOpenApi(_esi_data, endpoints=_endpoints)
esi_client_error_stub = EsiClientStub(_esi_data, endpoints=_endpoints, http_error=502)



# Generated by Django 4.2.11 on 2025-01-13 11:35

# Django
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("ledger", "0006_characteraudit_character_name_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="characterplanetdetails",
            name="facilitys",
            field=models.JSONField(blank=True, default=None, null=True),
        ),
    ]



# Standard Library
from unittest.mock import MagicMock, patch

# Django
from django.test import override_settings
from django.utils import timezone

# Alliance Auth (External Libs)
from app_utils.testing import NoSocketsTestCase

# AA Ledger
from ledger.models.general import EveEntity
from ledger.tests.testdata.esi_stub import esi_client_stub_openapi
from ledger.tests.testdata.generate_corporationaudit import (
    create_corporationaudit_from_user,
    create_user_from_evecharacter,
)
from ledger.tests.testdata.generate_walletjournal import (
    create_division,
    create_wallet_journal_entry,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveentity import load_eveentity
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "ledger.managers.corporation_journal_manager"


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODULE_PATH + ".esi")
@patch("ledger.models.general.EveEntity")
class TestCharacterJournalManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
        )
        cls.audit = create_corporationaudit_from_user(cls.user)

        cls.eve_character_first_party = EveEntity.objects.get(eve_id=2001)
        cls.eve_character_second_party = EveEntity.objects.get(eve_id=1001)

        cls.division = create_division(
            corporation=cls.audit, name="MEGA KONTO", balance=1000000, division_id=1
        )

        cls.journal_entry = create_wallet_journal_entry(
            journal_type="corporation",
            division=cls.division,
            context_id=1,
            entry_id=10,
            amount=1000,
            balance=2000,
            date=timezone.datetime.replace(
                timezone.now(),
                year=2016,
                month=10,
                day=29,
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ),
            description="Test Journal",
            first_party=cls.eve_character_first_party,
            second_party=cls.eve_character_second_party,
            ref_type="player_donation",
        )
        cls.token = cls.character_ownership.user.token_set.first()
        cls.audit.get_token = MagicMock(return_value=cls.token)

    def test_update_wallet_journal(self, mock_eveentity, mock_esi):
        # given
        mock_esi.client = esi_client_stub_openapi

        mock_eveentity.objects.create_bulk_from_esi.return_value = True

        EveEntity.objects.create(
            eve_id=9998, name="Test Character 2", category="character"
        )

        self.audit.update_wallet_journal(force_refresh=False)

        self.assertSetEqual(
            set(
                self.division.ledger_corporation_journal.values_list(
                    "entry_id", flat=True
                )
            ),
            {10, 13, 16},
        )
        obj = self.division.ledger_corporation_journal.get(entry_id=10)
        self.assertEqual(obj.amount, 1000)
        self.assertEqual(obj.context_id, 1)
        self.assertEqual(obj.first_party.eve_id, 2001)
        self.assertEqual(obj.second_party.eve_id, 1001)

        obj = self.division.ledger_corporation_journal.get(entry_id=13)
        self.assertEqual(obj.amount, 5000)

        obj = self.division.ledger_corporation_journal.get(entry_id=16)
        self.assertEqual(obj.amount, 10000)


class TestCorporationJournalManagerAnnotations(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
        )
        cls.audit = create_corporationaudit_from_user(cls.user)

        cls.eve_character_first_party = EveEntity.objects.get(eve_id=2001)
        cls.eve_character_second_party = EveEntity.objects.get(eve_id=1001)

        cls.division = create_division(
            corporation=cls.audit, name="MEGA KONTO", balance=1000000, division_id=1
        )

        cls.journal_entry = create_wallet_journal_entry(
            journal_type="corporation",
            division=cls.division,
            context_id=1,
            entry_id=10,
            amount=1000,
            balance=2000,
            date=timezone.datetime.replace(
                timezone.now(),
                year=2016,
                month=10,
                day=29,
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ),
            description="Test Journal",
            first_party=cls.eve_character_first_party,
            second_party=cls.eve_character_second_party,
            ref_type="player_donation",
        )

    def test_annotate_bounty_income(self):
        qs = self.division.ledger_corporation_journal.annotate_bounty_income()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "bounty_income"),
                "Bounty income annotation should be present",
            )
            self.assertEqual(obj.bounty_income, 0)

    def test_annotate_ess_income(self):
        qs = self.division.ledger_corporation_journal.annotate_ess_income()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "ess_income"),
                "ESS income annotation should be present",
            )
            self.assertEqual(obj.ess_income, 0)

    def test_annotate_miscellaneous(self):
        qs = self.division.ledger_corporation_journal.annotate_miscellaneous()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "miscellaneous"),
                "Miscellaneous annotation should be present",
            )
            self.assertEqual(obj.miscellaneous, 1000.00)

    def test_annotate_costs(self):
        qs = self.division.ledger_corporation_journal.annotate_costs()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "costs"),
                "Costs annotation should be present",
            )
            self.assertEqual(obj.costs, 0)


class TestCorporationJournalManagerAggregate(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
        )
        cls.audit = create_corporationaudit_from_user(cls.user)

        cls.eve_character_first_party = EveEntity.objects.get(eve_id=2001)
        cls.eve_character_second_party = EveEntity.objects.get(eve_id=1001)

        cls.division = create_division(
            corporation=cls.audit, name="MEGA KONTO", balance=1000000, division_id=1
        )

        cls.journal_entry = create_wallet_journal_entry(
            journal_type="corporation",
            division=cls.division,
            context_id=1,
            entry_id=10,
            amount=1000,
            balance=2000,
            date=timezone.datetime.replace(
                timezone.now(),
                year=2016,
                month=10,
                day=29,
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ),
            description="Test Journal",
            first_party=cls.eve_character_first_party,
            second_party=cls.eve_character_second_party,
            ref_type="player_donation",
        )

    def test_aggregate_bounty(self):
        result = self.division.ledger_corporation_journal.aggregate_bounty()
        self.assertEqual(result, 0)

    def test_aggregate_costs(self):
        result = self.division.ledger_corporation_journal.aggregate_costs()
        self.assertEqual(result, 0)

    def test_aggregate_miscellaneous(self):
        result = self.division.ledger_corporation_journal.aggregate_miscellaneous()
        self.assertEqual(result, 1000.00)

    def test_aggregate_ref_type(self):
        result = self.division.ledger_corporation_journal.aggregate_ref_type(
            ref_type=["player_donation"], income=True
        )
        self.assertEqual(result, 1000.00)



"""
General Model
"""

# Standard Library
import datetime
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, NamedTuple

# Third Party
from aiopenapi3.errors import ContentTypeError, HTTPClientError, HTTPServerError

# Django
from django.core.validators import MinValueValidator
from django.db import models
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)
from allianceauth.services.hooks import get_extension_logger
from esi.errors import TokenError
from esi.exceptions import HTTPNotModified

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__, app_settings
from ledger.managers.general_manager import EveEntityManager

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

# Permission Manager


class General(models.Model):
    """A model defining commonly used properties and methods for Ledger."""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app, Ledger."),
            ("advanced_access", "Can access Corporation and Alliance Ledger."),
            ("manage_access", "Can manage Ledger."),
        )


# EvE Entity Model - Store all Chars, Corps, Allys
class EveEntity(models.Model):
    """An Eve entity like a corporation or a character"""

    CATEGORY_ALLIANCE = "alliance"
    CATEGORY_CHARACTER = "character"
    CATEGORY_CORPORATION = "corporation"

    CATEGORY_CHOICES = (
        (CATEGORY_ALLIANCE, "Alliance"),
        (CATEGORY_CORPORATION, "Corporation"),
        (CATEGORY_CHARACTER, "Character"),
    )

    eve_id = models.IntegerField(
        primary_key=True, validators=[MinValueValidator(0)], verbose_name=_("ID")
    )
    category = models.CharField(
        max_length=32, choices=CATEGORY_CHOICES, verbose_name=_("Category")
    )
    name = models.CharField(max_length=254, verbose_name=_("Name"))

    # optionals for character/corp
    corporation = models.ForeignKey(
        "EveEntity",
        on_delete=models.SET_NULL,
        null=True,
        default=None,
        related_name="corp",
    )
    alliance = models.ForeignKey(
        "EveEntity",
        on_delete=models.SET_NULL,
        null=True,
        default=None,
        related_name="alli",
    )
    last_update = models.DateTimeField(auto_now=True)

    objects = EveEntityManager()

    def __str__(self) -> str:
        return str(self.name)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(id={self.eve_id}, category='{self.category}', "
            f"name='{self.name}')"
        )

    @property
    def is_alliance(self) -> bool:
        """Return True if entity is an alliance, else False."""
        return self.category == self.CATEGORY_ALLIANCE

    @property
    def is_corporation(self) -> bool:
        """Return True if entity is an corporation, else False."""
        return self.category == self.CATEGORY_CORPORATION

    @property
    def is_character(self) -> bool:
        """Return True if entity is a character, else False."""
        return self.category == self.CATEGORY_CHARACTER

    def icon_url(self, size=128) -> str:
        """Url to an icon image for this organization."""
        if self.category == self.CATEGORY_ALLIANCE:
            return EveAllianceInfo.generic_logo_url(self.eve_id, size=size)

        if self.category == self.CATEGORY_CORPORATION:
            return EveCorporationInfo.generic_logo_url(self.eve_id, size=size)

        if self.category == self.CATEGORY_CHARACTER:
            return EveCharacter.generic_portrait_url(self.eve_id, size=size)

        raise NotImplementedError(
            f"Avatar URL not implemented for category {self.category}"
        )

    def needs_update(self):
        return self.last_update + datetime.timedelta(days=7) < timezone.now()

    class Meta:
        default_permissions = ()


class UpdateSectionResult(NamedTuple):
    """A result of an attempted section update."""

    is_changed: bool | None
    is_updated: bool
    has_token_error: bool = False
    error_message: str | None = None
    data: Any = None


@dataclass(frozen=True)
class _NeedsUpdate:
    """An Object to track if an update is needed."""

    section_map: dict[str, bool]

    def __bool__(self) -> bool:
        """Check if any section needs an update."""
        return any(self.section_map.values())

    def for_section(self, section: str) -> bool:
        """Check if an update is needed for a specific section."""
        return self.section_map.get(section, False)


class AuditBase(models.Model):
    """A base model for audit models."""

    class Meta:
        abstract = True

    class UpdateStatus(models.TextChoices):
        DISABLED = "disabled", _("disabled")
        TOKEN_ERROR = "token_error", _("token error")
        ERROR = "error", _("error")
        OK = "ok", _("ok")
        INCOMPLETE = "incomplete", _("incomplete")
        IN_PROGRESS = "in_progress", _("in progress")

        def bootstrap_icon(self) -> str:
            """Return bootstrap corresponding icon class."""
            update_map = {
                status: mark_safe(
                    f"<span class='{self.bootstrap_text_style_class()}' data-tooltip-toggle='ledger-tooltip' title='{self.description()}'>⬤</span>"
                )
                for status in [
                    self.DISABLED,
                    self.TOKEN_ERROR,
                    self.ERROR,
                    self.INCOMPLETE,
                    self.IN_PROGRESS,
                    self.OK,
                ]
            }
            return update_map.get(self, "")

        def bootstrap_text_style_class(self) -> str:
            """Return bootstrap corresponding bootstrap text style class."""
            update_map = {
                self.DISABLED: "text-muted",
                self.TOKEN_ERROR: "text-warning",
                self.INCOMPLETE: "text-warning",
                self.IN_PROGRESS: "text-info",
                self.ERROR: "text-danger",
                self.OK: "text-success",
            }
            return update_map.get(self, "")

        def description(self) -> str:
            """Return description for an enum object."""
            update_map = {
                self.DISABLED: _("Update is disabled"),
                self.TOKEN_ERROR: _("One section has a token error during update"),
                self.INCOMPLETE: _("One or more sections have not been updated"),
                self.IN_PROGRESS: _("Update is in progress"),
                self.ERROR: _("An error occurred during update"),
                self.OK: _("Updates completed successfully"),
            }
            return update_map.get(self, "")

    def update_section_if_changed(
        self,
        section: models.TextChoices,
        fetch_func: Callable,
        force_refresh: bool = False,
    ):
        """Update the status of a specific section if it has changed."""
        section = self.UpdateSection(section)
        try:
            data = fetch_func(audit=self, force_refresh=force_refresh)
            logger.debug("%s: Update has changed, section: %s", self, section.label)
        except HTTPServerError as exc:
            logger.debug("%s: Update has an HTTP internal server error: %s", self, exc)
            return UpdateSectionResult(is_changed=False, is_updated=False)
        except HTTPClientError as exc:
            error_message = f"{type(exc).__name__}: {str(exc)}"
            # TODO ADD DISCORD/AUTH NOTIFICATION?
            logger.error(
                "%s: %s: Update has Client Error: %s %s",
                self,
                section.label,
                error_message,
                exc.status_code,
            )
            return UpdateSectionResult(
                is_changed=False,
                is_updated=False,
                has_token_error=True,
                error_message=error_message,
            )
        except HTTPNotModified:
            logger.debug("%s: Update has not changed, section: %s", self, section.label)
            return UpdateSectionResult(is_changed=False, is_updated=False)
        except (OSError, ContentTypeError) as exc:
            logger.info(
                "%s Update has a %s error, section: %s: %s",
                self,
                type(exc).__name__,
                section.label,
                exc,
            )
            return UpdateSectionResult(is_changed=False, is_updated=False)
        return UpdateSectionResult(
            is_changed=True,
            is_updated=True,
            data=data,
        )

    def reset_has_token_error(self) -> None:
        """Reset the has_token_error flag for this character."""
        if self.get_status == self.UpdateStatus.TOKEN_ERROR:
            self.update_status.filter(
                has_token_error=True,
            ).update(
                has_token_error=False,
            )
            return True
        return False

    def reset_update_status(self, section):
        """Reset the status of a given update section and return it."""
        update_status_obj = self.update_status.get_or_create(
            section=section,
        )[0]
        update_status_obj.reset()
        return update_status_obj

    def perform_update_status(
        self, section, method: Callable, *args, **kwargs
    ) -> UpdateSectionResult:
        """Perform update status."""
        try:
            result = method(*args, **kwargs)
        except Exception as exc:
            error_message = f"{type(exc).__name__}: {str(exc)}"
            is_token_error = isinstance(exc, (TokenError))
            logger.error(
                "%s: %s: Error during update status: %s",
                self,
                section.label,
                error_message,
                exc_info=not is_token_error,  # do not log token errors
            )

            # Update the status using the related manager name
            self.update_status.update_or_create(
                section=section,
                defaults={
                    "is_success": False,
                    "error_message": error_message,
                    "has_token_error": is_token_error,
                    "last_update_at": timezone.now(),
                },
            )
            raise exc
        return result

    def update_section_log(
        self,
        section: models.TextChoices,
        result: UpdateSectionResult,
    ) -> None:
        """Update the status of a specific section."""
        error_message = result.error_message if result.error_message else ""
        is_success = not result.has_token_error
        defaults = {
            "is_success": is_success,
            "error_message": error_message,
            "has_token_error": result.has_token_error,
            "last_run_finished_at": timezone.now(),
        }
        obj: UpdateStatus = self.update_status.update_or_create(
            section=section,
            defaults=defaults,
        )[0]
        if result.is_updated:
            obj.last_update_at = obj.last_run_at
            obj.last_update_finished_at = timezone.now()
            obj.save()
        status = "successfully" if is_success else "with errors"
        logger.info("%s: %s Update run completed %s", self, section.label, status)

    def calc_update_needed(self) -> _NeedsUpdate:
        """Calculate if an update is needed."""
        sections_needs_update = {
            section: True for section in self.UpdateSection.get_sections()
        }
        existing_sections: models.QuerySet[UpdateStatus] = self.update_status.all()
        needs_update = {
            obj.section: obj.need_update()
            for obj in existing_sections
            if obj.section in sections_needs_update
        }
        sections_needs_update.update(needs_update)
        return _NeedsUpdate(section_map=sections_needs_update)


class UpdateStatus(models.Model):
    """A Model to track the status of the last update."""

    is_success = models.BooleanField(default=None, null=True, db_index=True)
    error_message = models.TextField()
    has_token_error = models.BooleanField(default=False)

    last_run_at = models.DateTimeField(
        default=None,
        null=True,
        db_index=True,
        help_text="Last run has been started at this time",
    )
    last_run_finished_at = models.DateTimeField(
        default=None,
        null=True,
        db_index=True,
        help_text="Last run has been successful finished at this time",
    )
    last_update_at = models.DateTimeField(
        default=None,
        null=True,
        db_index=True,
        help_text="Last update has been started at this time",
    )
    last_update_finished_at = models.DateTimeField(
        default=None,
        null=True,
        db_index=True,
        help_text="Last update has been successful finished at this time",
    )

    class Meta:
        abstract = True
        default_permissions = ()

    def need_update(self) -> bool:
        """Check if the update is needed."""
        if not self.is_success or not self.last_update_finished_at:
            needs_update = True
        else:
            section_time_stale = app_settings.LEDGER_STALE_TYPES.get(self.section, 60)
            stale = timezone.now() - timezone.timedelta(minutes=section_time_stale)

            try:
                needs_update = self.last_update_finished_at <= stale
            except AttributeError:
                needs_update = True

        if needs_update and self.has_token_error:
            logger.info(
                "%s: Ignoring update because of token error, section: %s",
                self,
                self.section,
            )
            needs_update = False

        return needs_update

    def reset(self) -> None:
        """Reset this update status."""
        self.is_success = None
        self.error_message = ""
        self.has_token_error = False
        self.last_run_at = timezone.now()
        self.last_run_finished_at = None
        self.save()



# Django
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.evelinks.eveimageserver import (
    type_render_url,
)


def get_facilities_info(planet):
    facilities = []
    for __, facility_info in planet.facilitys.items():
        resources = facility_info["resources"]
        resource_counts = get_resource_counts(resources)

        input_icons = [
            {
                "icon_url": resource["item_id"],
                "item_name": resource["item_name"],
                "count": resource["count"],
            }
            for resource in resource_counts.values()
        ]

        output_icon = (
            {
                "icon_url": facility_info["output_product"]["item_id"],
                "item_name": facility_info["output_product"]["item_name"],
            }
            if facility_info["output_product"]
            else None
        )

        producing = any(resource["still_producing"] for resource in resources)
        is_active = not any(resource["missing_quantity"] > 0 for resource in resources)
        if producing and any(
            resource["missing_quantity"] > 0 for resource in resources
        ):
            is_active = True

        facilities.append(
            {
                "facility_name": facility_info["facility_name"] or _("No facility"),
                "input_icons": input_icons,
                "output_icon": output_icon,
                "is_active": is_active,
            }
        )
    return facilities


def get_resource_counts(resources):
    resource_counts = {}
    for resource in resources:
        if resource["item_id"] not in resource_counts:
            resource_counts[resource["item_id"]] = {
                "item_id": resource["item_id"],
                "item_name": resource["item_name"],
                "count": 0,
            }
        resource_counts[resource["item_id"]]["count"] += 1
    return resource_counts


def generate_progressbar(percentage: float) -> str:
    """
    Generate a progress bar based on the percentage.
    """
    percentage = round(percentage, 2)

    if percentage > 50:
        progress_value = format_html('<span class="text-white)">{}%</span>', percentage)
    else:
        progress_value = format_html('<span class="text-dark">{}%</span>', percentage)

    progressbar = format_html(
        """
        <div class="progress-outer flex-grow-1 me-2">
            <div class="progress" style="position: relative;">
                <div class="progress-bar progress-bar-warning progress-bar-striped active" role="progressbar" style="width: {}%; box-shadow: -1px 3px 5px rgba(0, 180, 231, 0.9);"></div>
                <div class="fw-bold fs-6 text-center position-absolute top-50 start-50 translate-middle">{}</div>
            </div>
        </div>
        """,
        percentage,
        progress_value,
    )

    return progressbar


def get_type_render_url(
    type_id: int, size: int = 32, type_name: str = None, as_html: bool = False
) -> str:
    """Get the type render for a type ID."""

    render_url = type_render_url(type_id=type_id, size=size)

    if as_html:
        render_html = format_html(
            '<img class="type-render rounded-circle" src="{}" title="{}">',
            render_url,
            type_name,
        )
        return render_html
    return render_url



# Standard Library
from unittest.mock import MagicMock

# Django
from django.test import TestCase

# AA Ledger
from ledger.templatetags.ledger import get_item, month_days, month_name, range_filter


class TestTemplateTags(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

    def test_month_name(self):
        # Test for valid month number
        self.assertEqual(month_name(1), "January")
        self.assertEqual(month_name(12), "December")

        # Test for invalid month number
        self.assertEqual(month_name(0), "")
        self.assertEqual(month_name(15), "")

    def test_month_days(self):
        # Test for a valid date
        date_info = {"year": 2023, "month": 2}
        self.assertEqual(month_days(date_info), list(range(1, 29)))

        # Test for a date with missing year
        date_info = {"month": 2}
        self.assertEqual(month_days(date_info), list(range(1, 29)))

        # Test for a date with missing month (approximate check)
        date_info = {"year": 2023}
        days = month_days(date_info)
        self.assertTrue(len(days) >= 28 and len(days) <= 31)
        self.assertEqual(days[0], 1)
        self.assertTrue(days[-1] >= 28 and days[-1] <= 31)

    def test_range_filter(self):
        # Test for a valid value
        self.assertEqual(list(range_filter(5)), [1, 2, 3, 4, 5])

        # Test for zero value
        self.assertEqual(list(range_filter(0)), [])

        # Test for negative value
        self.assertEqual(list(range_filter(-3)), [])

    def test_get_item(self):
        # Test for valid key
        data = {"key": "value"}
        self.assertEqual(get_item(data, "key"), "value")

        # Test for invalid key
        self.assertEqual(get_item(data, "invalid_key"), {})

        # Test for non-dictionary input
        self.assertEqual(get_item("not a dict", "key"), {})



# Standard Library
import logging
from unittest.mock import patch

# Django
from django.test import TestCase
from django.test.utils import override_settings
from django.utils import timezone

# AA Ledger
from ledger import tasks
from ledger.app_settings import LEDGER_CACHE_KEY
from ledger.models.planetary import CharacterPlanetDetails
from ledger.tests.testdata.generate_characteraudit import (
    create_characteraudit_from_evecharacter,
    create_update_status,
)
from ledger.tests.testdata.generate_corporationaudit import (
    create_corporation_update_status,
    create_corporationaudit_from_evecharacter,
)
from ledger.tests.testdata.generate_planets import (
    _planetary_data,
    create_character_planet,
    create_character_planet_details,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse

TASK_PATH = "ledger.tasks"


@patch(TASK_PATH + ".update_character", spec=True)
class TestUpdateAllCharacters(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()

        cls.audit = create_characteraudit_from_evecharacter(1001)

    def test_should_update_all_characters(self, mock_update_character):
        # when
        tasks.update_all_characters()
        # then
        self.assertTrue(mock_update_character.apply_async.called)

    def test_should_update_subset_characters(self, mock_update_character):
        # when
        tasks.update_subset_characters()
        # then
        self.assertTrue(mock_update_character.apply_async.called)


@patch(TASK_PATH + ".chain", spec=True)
@patch(TASK_PATH + ".logger", spec=True)
class TestUpdateCharacter(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()

        cls.audit = create_characteraudit_from_evecharacter(1001)

    def test_update_character_should_no_updated(self, mock_logger, __):
        # given
        for section in self.audit.UpdateSection.get_sections():
            create_update_status(
                self.audit,
                section=section,
                is_success=True,
                error_message="",
                has_token_error=False,
                last_run_at=timezone.now(),
                last_run_finished_at=timezone.now(),
                last_update_at=timezone.now(),
                last_update_finished_at=timezone.now(),
            )

        # when
        tasks.update_character(self.audit.pk)
        # then
        mock_logger.info.assert_called_once_with(
            "No updates needed for %s",
            self.audit.eve_character.character_name,
        )

    def test_update_character_should_update(self, mock_logger, mock_chain):
        # given
        create_update_status(
            self.audit,
            section=self.audit.UpdateSection.WALLET_JOURNAL,
            is_success=True,
            error_message="",
            has_token_error=False,
            last_run_at=None,
            last_run_finished_at=None,
            last_update_at=None,
            last_update_finished_at=None,
        )

        # when
        tasks.update_character(self.audit.pk)
        # then
        mock_chain.assert_called_once()


@patch(TASK_PATH + ".update_corporation", spec=True)
class TestUpdateAllCorporations(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()

        cls.audit = create_corporationaudit_from_evecharacter(1001)

    def test_should_update_all_corporations(self, mock_update_corporation):
        # when
        tasks.update_all_corporations()
        # then
        self.assertTrue(mock_update_corporation.apply_async.called)

    def test_should_update_subset_corporation(self, mock_update_corporation):
        # when
        tasks.update_subset_corporations()
        # then
        self.assertTrue(mock_update_corporation.apply_async.called)


@patch(TASK_PATH + ".chain", spec=True)
@patch(TASK_PATH + ".logger", spec=True)
class TestUpdateCorporation(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()

        cls.audit = create_corporationaudit_from_evecharacter(1001)

    def test_update_corporation_should_no_updated(self, mock_logger, __):
        # given
        for section in self.audit.UpdateSection.get_sections():
            create_corporation_update_status(
                self.audit,
                section=section,
                is_success=True,
                error_message="",
                has_token_error=False,
                last_run_at=timezone.now(),
                last_run_finished_at=timezone.now(),
                last_update_at=timezone.now(),
                last_update_finished_at=timezone.now(),
            )
        # when
        tasks.update_corporation(self.audit.pk)
        # then
        mock_logger.info.assert_called_once_with(
            "No updates needed for %s",
            self.audit.corporation.corporation_name,
        )

    def test_update_corporation_should_update(self, mock_logger, mock_chain):
        # given
        create_corporation_update_status(
            self.audit,
            section=self.audit.UpdateSection.WALLET_JOURNAL,
            is_success=True,
            error_message="",
            has_token_error=False,
            last_run_at=None,
            last_run_finished_at=None,
            last_update_at=None,
            last_update_finished_at=None,
        )

        # when
        tasks.update_corporation(self.audit.pk)
        # then
        mock_chain.assert_called_once()


@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
@patch(TASK_PATH + ".logger", spec=True)
class TestClearEtag(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()

    def test_clear_all_etag_return_no_etags(self, mock_logger):
        # when
        tasks.clear_all_etags()

        # then
        mock_logger.info.assert_any_call("Deleting %s etag keys", 0)
        mock_logger.info.assert_any_call("No etag keys to delete")

    def test_clear_all_etag_return_etags(self, mock_logger):
        # given
        # pylint: disable=import-outside-toplevel
        # Third Party
        from django_redis import get_redis_connection

        _client = get_redis_connection("default")
        keys = _client.keys(f":?:{LEDGER_CACHE_KEY}-*")

        # when
        tasks.clear_all_etags()

        # then
        mock_logger.info.assert_any_call("Deleting %s etag keys", len(keys))
        mock_logger.info.assert_any_call("Deleted %s etag keys", len(keys))


@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
class TestCheckPlanetaryNotification(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.planet_params = {
            "upgrade_level": 5,
            "num_pins": 5,
        }

        cls.planetdetails_params = {
            "pin_id": 1,
            "pin_type": "Extractor",
            "status": "Active",
            "last_update": None,
            "last_update_finished_at": None,
        }

        cls.audit = create_characteraudit_from_evecharacter(1001)

        cls.audit2 = create_characteraudit_from_evecharacter(1002)

        cls.planet_1 = create_character_planet(
            cls.audit, planet_id=4001, **cls.planet_params
        )

        cls.planet_2 = create_character_planet(
            cls.audit2, planet_id=4001, **cls.planet_params
        )

        cls.planet_3 = create_character_planet(
            cls.audit2, planet_id=4002, **cls.planet_params
        )

        cls.planetdetails_1 = create_character_planet_details(
            cls.planet_1, notification=True, **_planetary_data
        )

        cls.planetdetails_2 = create_character_planet_details(
            cls.planet_2,
            notification=True,
        )

        cls.planetdetails_3 = create_character_planet_details(
            cls.planet_3, notification=True, **_planetary_data
        )

    def test_check_planetary_notification(self):
        # when
        tasks.check_planetary_alarms()
        # then
        self.assertEqual(
            CharacterPlanetDetails.objects.filter(
                notification_sent=True,
            ).count(),
            2,
        )



"""TestView class."""

# Standard Library
from http import HTTPStatus
from unittest.mock import Mock, patch

# Django
from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase, override_settings
from django.urls import reverse
from django.utils import timezone

# AA Ledger
from ledger.models.general import EveEntity
from ledger.tests.testdata.generate_corporationaudit import (
    create_corporationaudit_from_user,
    create_user_from_evecharacter,
)
from ledger.tests.testdata.generate_walletjournal import (
    create_division,
    create_wallet_journal_entry,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveentity import load_eveentity
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse
from ledger.views.corporation.corporation_ledger import (
    corporation_data_export_generate,
    corporation_data_export_run_update,
)

MODULE_PATH = "ledger.views.corporation.corporation_ledger"


@patch(MODULE_PATH + ".messages")
@patch(MODULE_PATH + ".tasks")
@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestCorporationLedgerView(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.advanced_access",
                "ledger.manage_access",
            ],
        )

        cls.user_no_permissions, cls.character_ownership = (
            create_user_from_evecharacter(
                1002,
                permissions=[
                    "ledger.basic_access",
                    "ledger.advanced_access",
                    "ledger.manage_access",
                ],
            )
        )
        cls.audit = create_corporationaudit_from_user(cls.user)

        cls.eve_character_first_party = EveEntity.objects.get(eve_id=2001)
        cls.eve_character_second_party = EveEntity.objects.get(eve_id=1001)

        cls.division = create_division(
            corporation=cls.audit, name="MEGA KONTO", balance=1000000, division_id=1
        )

        cls.journal_entry = create_wallet_journal_entry(
            journal_type="corporation",
            division=cls.division,
            context_id=1,
            entry_id=10,
            amount=1000,
            balance=2000,
            date=timezone.datetime.replace(
                timezone.now(),
                year=2016,
                month=10,
                day=29,
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ),
            description="Test Journal",
            first_party=cls.eve_character_first_party,
            second_party=cls.eve_character_second_party,
            ref_type="player_donation",
        )

    def test_corporation_data_export_generate(self, mock_tasks, mock_messages):
        """Test should generate corporation data export"""
        corporation_id = self.audit.corporation.corporation_id
        mock_tasks.export_data_ledger.apply_async = Mock()

        form_data = {
            "year": "2016",
        }

        request = self.factory.post(
            reverse("ledger:corporation_data_export_generate", args=[corporation_id]),
            data=form_data,
        )

        request.user = self.user

        response = corporation_data_export_generate(
            request, corporation_id=corporation_id
        )

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        mock_messages.info.assert_called_once_with(
            request,
            f"Data export for {corporation_id} has been started. This can take a couple of minutes. You will get a notification once it is completed.",
        )
        mock_tasks.export_data_ledger.apply_async.assert_called_once()

    def test_corporation_data_export_generate_no_permission(
        self, mock_tasks, mock_messages
    ):
        """Test should not generate corporation data export due to no permission"""
        corporation_id = self.audit.corporation.corporation_id
        mock_tasks.export_data_ledger.apply_async = Mock()

        form_data = {
            "year": "2016",
        }

        request = self.factory.post(
            reverse("ledger:corporation_data_export_generate", args=[corporation_id]),
            data=form_data,
        )

        request.user = self.user_no_permissions

        response = corporation_data_export_generate(
            request, corporation_id=corporation_id
        )

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        mock_messages.error.assert_called_once_with(request, "Permission Denied")
        mock_tasks.export_data_ledger.apply_async.assert_not_called()

    def test_corporation_data_export_generate_corporation_not_found(
        self, mock_tasks, mock_messages
    ):
        """Test should not generate corporation data export due to corporation not found"""
        corporation_id = 9999  # Non-existent corporation ID
        mock_tasks.export_data_ledger.apply_async = Mock()

        form_data = {
            "year": "2016",
        }

        request = self.factory.post(
            reverse("ledger:corporation_data_export_generate", args=[corporation_id]),
            data=form_data,
        )

        request.user = self.user

        response = corporation_data_export_generate(
            request, corporation_id=corporation_id
        )

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        mock_messages.info.assert_called_once_with(request, "Corporation not found")
        mock_tasks.export_data_ledger.apply_async.assert_not_called()

    def test_corporation_data_export_generate_invalid_form(
        self, mock_tasks, mock_messages
    ):
        """Test should not generate corporation data export due to invalid form"""
        corporation_id = self.audit.corporation.corporation_id
        mock_tasks.export_data_ledger.apply_async = Mock()

        form_data = {
            "year": "invalid_year",  # Invalid year
        }

        request = self.factory.post(
            reverse("ledger:corporation_data_export_generate", args=[corporation_id]),
            data=form_data,
        )

        request.user = self.user

        response = corporation_data_export_generate(
            request, corporation_id=corporation_id
        )

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        mock_messages.error.assert_called_once_with(request, "Invalid form submission.")
        mock_tasks.export_data_ledger.apply_async.assert_not_called()

    @patch(MODULE_PATH + ".data_exporter.LedgerCSVExporter.decoder")
    def test_corporation_data_export_run_update(
        self, mock_decoder, mock_tasks, mock_messages
    ):
        """Test should run corporation data export update task"""
        corporation_id = self.audit.corporation.corporation_id
        hashcode = "testhashcode"
        mock_tasks.export_data_ledger.apply_async = Mock()
        mock_decoder.return_value = (2001, 1, 2016, 1)

        request = self.factory.get(
            reverse("ledger:corporation_data_export_run_update", args=[hashcode]),
        )

        request.user = self.user

        response = corporation_data_export_run_update(request, hash_code=hashcode)

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        mock_messages.info.assert_called_once_with(
            request,
            f"Data export for {corporation_id} has been started. This can take a couple of minutes. You will get a notification once it is completed.",
        )
        mock_tasks.export_data_ledger.apply_async.assert_called_once()

    @patch(MODULE_PATH + ".data_exporter.LedgerCSVExporter.decoder")
    def test_corporation_data_export_run_update_no_permission(
        self, mock_decoder, mock_tasks, mock_messages
    ):
        """Test should not run corporation data export update task due to no permission"""
        hashcode = "testhashcode"
        mock_tasks.export_data_ledger.apply_async = Mock()
        mock_decoder.return_value = (2001, 1, 2016, 1)

        request = self.factory.get(
            reverse("ledger:corporation_data_export_run_update", args=[hashcode]),
        )

        request.user = self.user_no_permissions

        response = corporation_data_export_run_update(request, hash_code=hashcode)

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        mock_messages.error.assert_called_once_with(request, "Permission Denied")
        mock_tasks.export_data_ledger.apply_async.assert_not_called()

    @patch(MODULE_PATH + ".data_exporter.LedgerCSVExporter.decoder")
    def test_corporation_data_export_run_update_corporation_not_found(
        self, mock_decoder, mock_tasks, mock_messages
    ):
        """Test should not run corporation data export update task due to corporation not found"""
        hashcode = "testhashcode"
        mock_tasks.export_data_ledger.apply_async = Mock()
        mock_decoder.return_value = (9999, 1, 2016, 1)

        request = self.factory.get(
            reverse("ledger:corporation_data_export_run_update", args=[hashcode]),
        )

        request.user = self.user

        response = corporation_data_export_run_update(request, hash_code=hashcode)

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        mock_messages.info.assert_called_once_with(request, "Corporation not found")
        mock_tasks.export_data_ledger.apply_async.assert_not_called()



# Generated by Django 4.2.23 on 2025-07-21 18:21

# Django
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("ledger", "0013_delete_events"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="corporationaudit",
            options={
                "default_permissions": (),
                "permissions": (
                    ("corp_audit_manager", "Has Access to own Corporations."),
                    ("corp_audit_admin_manager", "Has access to all Corporations."),
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="general",
            options={
                "default_permissions": (),
                "managed": False,
                "permissions": (
                    ("basic_access", "Can access this app, Ledger."),
                    ("advanced_access", "Can access Corporation and Alliance Ledger."),
                    ("manage_access", "Can manage Ledger."),
                ),
            },
        ),
        migrations.AlterField(
            model_name="corporationupdatestatus",
            name="section",
            field=models.CharField(
                choices=[
                    ("wallet_division_names", "Divisions Names"),
                    ("wallet_division", "Divisions"),
                    ("wallet_journal", "Wallet Journal"),
                ],
                db_index=True,
                max_length=32,
            ),
        ),
    ]



"""PvE Views"""

# Standard Library
from decimal import Decimal
from typing import Any

# Django
from django.core.handlers.wsgi import WSGIRequest
from django.db.models import DecimalField, F, Q, QuerySet, Sum
from django.utils.translation import gettext as _

# Alliance Auth
from allianceauth.eveonline.models import EveAllianceInfo
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.helpers.billboard import BillboardSystem
from ledger.helpers.core import LedgerCore, LedgerEntity
from ledger.helpers.ref_type import RefTypeManager
from ledger.models.corporationaudit import (
    CorporationAudit,
    CorporationWalletJournalEntry,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class AllianceData(LedgerCore):
    """Class to hold alliance data for the ledger."""

    # pylint: disable=too-many-positional-arguments
    def __init__(
        self,
        alliance: EveAllianceInfo,
        request: WSGIRequest = None,
        year=None,
        month=None,
        day=None,
        section=None,
    ):
        super().__init__(year, month, day)
        self.request = request
        self.alliance = alliance
        self.entity_id = self.alliance.alliance_id
        self.section = section
        self.corporations = CorporationAudit.objects.filter(
            corporation__alliance__alliance_id=self.alliance.alliance_id
        ).values_list("corporation__corporation_id", flat=True)
        self.auth_char_ids = self.auth_character_ids
        self.billboard = BillboardSystem()
        self.queryset = (
            self._get_journal_queryset()
        )  # Base queryset filtered by date and alliance

    def _get_journal_queryset(self) -> QuerySet[CorporationWalletJournalEntry]:
        """Return the base queryset filtered by the current date range and corporation division."""
        return CorporationWalletJournalEntry.objects.filter(
            self.filter_date,
            division__corporation__corporation__alliance__alliance_id=self.alliance.alliance_id,
        ).exclude(
            Q(ref_type="corporation_account_withdrawal")
            & Q(first_party_id=F("second_party_id"))
        )

    def _compute_entities(
        self, journal: QuerySet[CorporationWalletJournalEntry]
    ) -> set:
        """Return a set of all entity IDs (first and second parties) present in the current journal."""
        return set(journal.values_list("second_party_id", flat=True)) | set(
            journal.values_list("first_party_id", flat=True)
        )

    def _compute_journal_values(
        self, journal: QuerySet[CorporationWalletJournalEntry]
    ) -> QuerySet[dict[str, Any]]:
        """Return the journal values for the current journal."""
        return journal.values(
            "first_party_id",
            "second_party_id",
            "pk",
            "ref_type",
            "division__corporation__corporation__corporation_id",
        ).annotate(
            bounty=Sum(
                "amount",
                filter=Q(ref_type__in=RefTypeManager.BOUNTY_PRIZES),
                output_field=DecimalField(),
            ),
            ess=Sum(
                "amount",
                filter=Q(ref_type__in=RefTypeManager.ESS_TRANSFER),
                output_field=DecimalField(),
            ),
            costs=Sum(
                "amount",
                filter=Q(ref_type__in=RefTypeManager.all_ref_types(), amount__lt=0),
                output_field=DecimalField(),
            ),
            miscellaneous=Sum(
                "amount",
                filter=Q(ref_type__in=RefTypeManager.all_ref_types(), amount__gt=0),
                output_field=DecimalField(),
            ),
        )

    # pylint: disable=duplicate-code
    def create_entity_data(
        self,
        entity: LedgerEntity,
    ) -> dict:
        """Create the URL for entity details based on the view type."""
        used_pks = set()
        bounty = Decimal(0)
        ess = Decimal(0)
        miscellaneous = Decimal(0)
        costs = Decimal(0)

        for pk, rows in list(self.entries.items()):
            for row in rows:
                if (
                    row["division__corporation__corporation__corporation_id"]
                    == entity.entity_id
                ):
                    if RefTypeManager.special_cases(
                        row, ids=[], account_char_ids=self.auth_char_ids
                    ):
                        continue
                    bounty += row.get("bounty") or Decimal(0)
                    ess += row.get("ess") or Decimal(0)
                    miscellaneous += row.get("miscellaneous") or Decimal(0)
                    costs += row.get("costs") or Decimal(0)
                    used_pks.add(pk)

        # Remove Used Pks from Entries
        # This is to prevent the entries from being used in the future
        for pk in used_pks:
            self.entries.pop(pk, None)

        misc = miscellaneous
        total = sum([bounty, ess, miscellaneous, costs])

        if total == 0:
            return None

        entity_ledger_info = {
            "entity": entity,
            "ledger": {
                "bounty": bounty,
                "ess": ess,
                "miscellaneous": misc,
                "costs": costs,
                "total": total,
            },
            "type": entity.type,
        }

        return entity_ledger_info

    # pylint: disable=duplicate-code
    def generate_ledger_data(self) -> dict:
        """Generate the ledger data for the alliance."""
        # Compute all entities in the journal
        self.entities = self._compute_entities(self.queryset)
        # Compute journal values
        journal = self._compute_journal_values(self.queryset)

        # Caching
        ledger_hash = self.get_ledger_journal_hash(journal.values_list("pk"))
        cache_key = f"{self.entity_id}"

        # Get Cached Data if available
        ledger, finished_entities = self.get_cache_ledger(
            ledger_hash=ledger_hash, cache_key=cache_key
        )

        if finished_entities is False or ledger is False:
            ledger = []
            finished_entities = set()

            # Build the entries from the journal
            self.entries = {}
            for row in journal:
                self.entries.setdefault(row["pk"], []).append(row)

            # Build Data for each corporation
            for corporation_id in self.corporations:
                # Create Details URL for the entity
                details_url = self.create_url(
                    viewname="corporation_details",
                    corporation_id=corporation_id,
                    entity_id=corporation_id,
                    section="summary",
                )

                # Create the LedgerEntity object for the entity
                entity_obj = LedgerEntity(
                    entity_id=corporation_id,
                    details_url=details_url,
                )

                corp_data = self.create_entity_data(
                    entity=entity_obj,
                )

                if corp_data is None:
                    continue

                ledger.append(corp_data)
                finished_entities.add(corporation_id)

            # Create Cache
            self.set_cache_ledger(
                ledger_hash=ledger_hash,
                cache_key=cache_key,
                ledger=ledger,
                finished_entities=finished_entities,
            )

        # Create the billboard data
        self.billboard.change_view(self.get_view_mode())
        self.create_rattingbar(journal=journal, entities_ids=list(finished_entities))
        self.create_chord(ledger)
        return ledger

    def generate_data_export(self) -> dict:
        """Generate the data export for the corporation."""
        # Compute all entities in the journal
        self.entities = self._compute_entities(self.queryset)
        # Compute journal values
        journal = self._compute_journal_values(self.queryset)

        ledger = []

        # Build the entries from the journal
        self.entries = {}
        for row in journal:
            self.entries.setdefault(row["pk"], []).append(row)

        # Build Data for each corporation
        for corporation_id in self.corporations:
            # Create the LedgerEntity object for the entity
            entity_obj = LedgerEntity(
                entity_id=corporation_id,
            )

            # Create Ledger Data for the entity
            corp_data = self.create_entity_data(
                entity=entity_obj,
            )

            if corp_data is None:
                continue

            ledger.append(corp_data)
        return ledger

    def create_rattingbar(
        self,
        journal: QuerySet[CorporationWalletJournalEntry],
        entities_ids: list = None,
    ):
        """Create the ratting bar for the view."""
        if not entities_ids:
            return

        # Create the timeline for the ratting bar
        rattingbar_timeline = self.billboard.create_timeline(journal)

        # Annotate the timeline with the relevant data
        rattingbar = (
            rattingbar_timeline.annotate_bounty_income()
            .annotate_ess_income()
            .annotate_miscellaneous()
        )

        # Generate the XY series for the ratting bar
        self.billboard.create_or_update_results(rattingbar)
        series, categories = self.billboard.generate_xy_series()
        if series and categories:
            # Create the ratting bar chart
            self.billboard.create_xy_chart(
                title=_("Ratting Bar"), categories=categories, series=series
            )

    def create_chord(self, ledger_data: list[dict]):
        """Create the chord chart for the view."""
        if not ledger_data:
            return

        for entry in ledger_data:
            entity_name = entry["entity"].entity_name
            ledger = entry["ledger"]
            self.billboard.chord_add_data(
                chord_from=entity_name,
                chord_to=_("Bounty (Wallet)"),
                value=ledger.get("bounty", 0),
            )
            self.billboard.chord_add_data(
                chord_from=entity_name,
                chord_to=_("ESS (Wallet)"),
                value=ledger.get("ess", 0),
            )
            self.billboard.chord_add_data(
                chord_from=entity_name,
                chord_to=_("Costs (Wallet)"),
                value=abs(ledger.get("costs", 0)),
            )
            self.billboard.chord_add_data(
                chord_from=entity_name,
                chord_to=_("Miscellaneous (Wallet)"),
                value=abs(ledger.get("miscellaneous", 0)),
            )
        self.billboard.chord_handle_overflow()



# Django
from django.contrib.auth.models import Permission
from django.test import TestCase

# Alliance Auth
from allianceauth.tests.auth_utils import AuthUtils

# AA Ledger
from ledger.models.corporationaudit import CorporationAudit
from ledger.tests.testdata.generate_corporationaudit import (
    create_corporationaudit_from_user,
    create_user_from_evecharacter,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth

MODULE_PATH = "ledger.models.corporationaudit"


class TestCorporationAuditModel(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001, permissions=["ledger.basic_access"]
        )
        cls.user2, cls.character_ownership2 = create_user_from_evecharacter(
            1002, permissions=["ledger.basic_access"]
        )
        cls.audit = create_corporationaudit_from_user(cls.user)
        cls.audit2 = create_corporationaudit_from_user(cls.user2)

    def test_str(self):
        expected_str = CorporationAudit.objects.get(id=self.audit.pk)
        self.assertEqual(self.audit, expected_str)

    def test_get_esi_scopes(self):
        self.assertEqual(
            self.audit.get_esi_scopes(),
            [
                # General
                "esi-search.search_structures.v1",
                "esi-universe.read_structures.v1",
                "esi-characters.read_corporation_roles.v1",
                # Mining
                "esi-industry.read_corporation_mining.v1",
                # wallets
                "esi-wallet.read_corporation_wallets.v1",
                "esi-markets.read_corporation_orders.v1",
                "esi-industry.read_corporation_jobs.v1",
                "esi-corporations.read_divisions.v1",
            ],
        )

    def test_access_no_perms(self):
        corporation = CorporationAudit.objects.visible_to(self.user)
        self.assertNotIn(self.audit, corporation)
        self.assertNotIn(self.audit2, corporation)

    def test_access_perms_own_corp(self):
        self.user = AuthUtils.add_permission_to_user_by_name(
            "ledger.advanced_access", self.user
        )
        self.user.refresh_from_db()
        corporation = CorporationAudit.objects.visible_to(self.user)
        self.assertIn(self.audit, corporation)
        self.assertNotIn(self.audit2, corporation)



"""PvE Views"""

# Standard Library
import json
from http import HTTPStatus

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.core.handlers.wsgi import WSGIRequest
from django.http import FileResponse, Http404, JsonResponse
from django.shortcuts import redirect, render
from django.utils.translation import gettext_lazy as _
from django.views.decorators.http import require_POST

# Alliance Auth
from allianceauth.authentication.models import CharacterOwnership
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__, forms, tasks
from ledger.api.helpers import (
    get_corporation,
    get_manage_corporation,
)
from ledger.helpers import data_exporter
from ledger.helpers.core import add_info_to_context
from ledger.helpers.corporation import CorporationData, LedgerEntity
from ledger.models.corporationaudit import (
    CorporationAudit,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("ledger.advanced_access")
def corporation_ledger_index(request):
    """Corporation Ledger Index View"""
    context = {}
    context = add_info_to_context(request, context)
    return redirect(
        "ledger:corporation_ledger", request.user.profile.main_character.corporation_id
    )


@login_required
@permission_required("ledger.basic_access")
def corporation_data_export(request, corporation_id: int):
    """Data Export View"""
    perms = get_corporation(request, corporation_id)[0]

    corporation_exporter = data_exporter.LedgerCSVExporter.create_exporter(
        "corporation", corporation_id
    )
    files = corporation_exporter.gather_export_files()

    context = {
        "title": "Data Export",
        "corporation_id": corporation_id,
        "files": files,
        "is_exportable": corporation_exporter.has_data,
    }

    if perms is False:
        msg = _("Permission Denied")
        messages.error(request, msg)
        return render(request, "ledger/data-export.html", context=context)

    if perms is None:
        msg = _("Corporation not found")
        messages.info(request, msg)
        return render(request, "ledger/data-export.html", context=context)

    context.update(
        {
            "forms": {
                "generate_data_export": forms.GenerateDataExportForm(
                    corporation_id=corporation_id
                ),
            },
        }
    )

    context = add_info_to_context(request, context)
    return render(request, "ledger/data-export.html", context)


@login_required
@permission_required("ledger.manage_access")
# pylint: disable=unused-argument
def corporation_download_export_file(
    request,
    hash_code: str,
) -> FileResponse:
    """Render file view for downloading an export file."""
    entity_id, division_id, year, month = data_exporter.LedgerCSVExporter.decoder(
        hash_code
    )
    exporter = data_exporter.LedgerCSVExporter.create_exporter(
        "corporation", entity_id, division_id=division_id, year=year, month=month
    )
    destination = data_exporter.default_destination()
    zip_file = destination / exporter.output_basename.with_suffix(".zip")
    if not zip_file.exists():
        raise Http404(f"Could not find export file for corporation {entity_id}")
    logger.info("Returning file %s for download of topic %s", zip_file, "corporation")
    return FileResponse(zip_file.open("rb"))


@login_required
@permission_required("ledger.manage_access")
@require_POST
def corporation_data_export_generate(request, corporation_id: int):
    """Handle POST form to generate a data export for a corporation."""
    perms = get_corporation(request, corporation_id)[0]
    if perms is False:
        msg = _("Permission Denied")
        messages.error(request, msg)
        return redirect("ledger:corporation_data_export", corporation_id=corporation_id)

    if perms is None:
        msg = _("Corporation not found")
        messages.info(request, msg)
        return redirect("ledger:corporation_data_export", corporation_id=corporation_id)

    form = forms.GenerateDataExportForm(
        request.POST,
        corporation_id=corporation_id,
    )

    if not form.is_valid():
        msg = _("Invalid form submission.")
        messages.error(request, msg)
        return redirect("ledger:corporation_data_export", corporation_id=corporation_id)

    # Read optional form values
    year_val = form.cleaned_data.get("year")
    month_val = form.cleaned_data.get("month") or None

    division_obj = form.cleaned_data.get("division")
    division_val = division_obj.division_id if division_obj is not None else None

    logger.debug("Generating data export for year=%s, month=%s", year_val, month_val)
    tasks.export_data_ledger.apply_async(
        kwargs={
            "user_pk": request.user.pk,
            "ledger_type": "corporation",
            "entity_id": corporation_id,
            "division_id": division_val,
            "year": year_val,
            "month": month_val,
        },
        priority=7,
    )
    msg = _(
        f"Data export for {corporation_id} has been started. This can take a couple of minutes. You will get a notification once it is completed."
    )
    messages.info(request, msg)
    return redirect("ledger:corporation_data_export", corporation_id=corporation_id)


@login_required
@permission_required("ledger.manage_access")
def corporation_data_export_run_update(
    request,
    hash_code: str,
):
    """Render view for running data export update."""
    entity_id, division_id, year, month = data_exporter.LedgerCSVExporter.decoder(
        hash_code
    )

    logger.debug(
        "Running data export update for entity_id=%s, division_id=%s, year=%s, month=%s",
        entity_id,
        division_id,
        year,
        month,
    )
    perms = get_corporation(request, entity_id)[0]
    if perms is False:
        msg = _("Permission Denied")
        messages.error(request, msg)
        return redirect("ledger:corporation_data_export", corporation_id=entity_id)

    if perms is None:
        msg = _("Corporation not found")
        messages.info(request, msg)
        return redirect("ledger:corporation_data_export", corporation_id=entity_id)

    tasks.export_data_ledger.apply_async(
        kwargs={
            "user_pk": request.user.pk,
            "ledger_type": "corporation",
            "entity_id": entity_id,
            "division_id": division_id,
            "year": year,
            "month": month,
        },
        priority=7,
    )
    msg = _(
        f"Data export for {entity_id} has been started. This can take a couple of minutes. You will get a notification once it is completed."
    )
    messages.info(request, msg)
    return redirect("ledger:corporation_data_export", corporation_id=entity_id)


# pylint: disable=too-many-positional-arguments
@login_required
@permission_required("ledger.advanced_access")
def corporation_ledger(
    request: WSGIRequest,
    corporation_id: int,
    division_id: int = None,
    year: int = None,
    month: int = None,
    day: int = None,
):
    """
    Corporation Ledger
    """
    perms, corporation = get_corporation(request, corporation_id)

    context = {
        "title": "Corporation Ledger",
        "corporation_id": corporation_id,
        "disabled": True,
    }

    # pylint: disable=duplicate-code
    if perms is False:
        msg = _("Permission Denied")
        messages.error(request, msg)
        return render(
            request, "ledger/corpledger/corporation_ledger.html", context=context
        )
    # pylint: disable=duplicate-code
    if perms is None:
        msg = _("Corporation not found")
        messages.info(request, msg)
        return render(
            request, "ledger/corpledger/corporation_ledger.html", context=context
        )

    corporation_data = CorporationData(
        corporation=corporation,
        division_id=division_id,
        year=year,
        month=month,
        day=day,
    )
    # Create the Corporation ledger data
    ledger = corporation_data.generate_ledger_data()

    context = {
        "title": f"Corporation Ledger - {corporation.corporation.corporation_name}",
        "corporation_id": corporation_id,
        "division_id": division_id,
        "billboard": json.dumps(corporation_data.billboard.dict.asdict()),
        "ledger": ledger,
        "divisions": corporation_data.divisions,
        "years": corporation_data.activity_years,
        "totals": corporation_data.calculate_totals(ledger),
        "view": corporation_data.create_view_data(
            viewname="corporation_details",
            corporation_id=corporation_id,
            entity_id=corporation_id,
            division_id=division_id,
            section="summary",
        ),
    }
    # Add additional information to the context
    context = add_info_to_context(request, context)

    return render(request, "ledger/corpledger/corporation_ledger.html", context=context)


# pylint: disable=too-many-positional-arguments, too-many-arguments
@login_required
@permission_required("ledger.advanced_access")
def corporation_details(
    request: WSGIRequest,
    corporation_id: int,
    entity_id: int,
    division_id: int = None,
    year: int = None,
    month: int = None,
    day: int = None,
    section: str = None,
):
    """
    Corporation Details
    """
    perms, corporation = get_corporation(request, corporation_id)

    context = {
        "title": "Corporation Ledger",
        "corporation_id": corporation_id,
    }

    # pylint: disable=duplicate-code
    if perms is False:
        msg = _("Permission Denied")
        return render(
            request,
            "ledger/partials/information/view_character_content.html",
            {
                "error": msg,
                "corporation_id": corporation_id,
            },
        )
    # pylint: disable=duplicate-code
    if perms is None:
        msg = _("Corporation not found")
        return render(
            request,
            "ledger/partials/information/view_character_content.html",
            {
                "error": msg,
                "corporation_id": corporation_id,
            },
        )

    corporation_data = CorporationData(
        corporation=corporation,
        division_id=division_id,
        year=year,
        month=month,
        day=day,
        section=section,
    )

    # Create the Entity for the ledger
    entity = LedgerEntity(
        entity_id=entity_id,
    )

    journal = corporation_data.filter_entity_journal(entity=entity)
    amounts = corporation_data._create_corporation_details(
        journal=journal, entity=entity
    )
    details = corporation_data._add_average_details(request, amounts, day)

    context = {
        "title": f"Corporation Details - {corporation.corporation_name}",
        "type": "corporation",
        "character": details,
        "information": f"Corporation Details - {corporation_data.get_details_title}",
    }
    context = add_info_to_context(request, context)
    # pylint: disable=duplicate-code
    return render(
        request,
        "ledger/partials/information/view_character_content.html",
        context=context,
    )


@login_required
@permission_required("ledger.advanced_access")
def corporation_overview(request):
    """
    Corporation Overview
    """
    context = {"title": "Corporation Overview"}
    context = add_info_to_context(request, context)
    return render(
        request, "ledger/corpledger/admin/corporation_overview.html", context=context
    )


@login_required
@permission_required("ledger.manage_access")
def corporation_administration(request, corporation_id):
    """
    Corporation Administration
    """
    perm, corporation = get_manage_corporation(request, corporation_id)

    if perm is False:
        msg = _("Permission Denied")
        messages.error(request, msg)
        return redirect("ledger:corporation_ledger_index")

    if perm is None:
        msg = _("Corporation not found")
        messages.info(request, msg)
        return redirect("ledger:corporation_ledger_index")

    # TODO Get Missing Characters from esi-corporations.read_corporation_membership.v1 ?
    corp_characters = CharacterOwnership.objects.filter(
        character__corporation_id=corporation.corporation.corporation_id
    ).order_by("character__character_name")
    corporation_dataexporter = data_exporter.LedgerCSVExporter.create_exporter(
        "corporation", corporation_id
    )

    context = {
        "corporation_id": corporation_id,
        "title": "Corporation Administration",
        "corporation": corporation,
        "characters": corp_characters,
        "is_exportable": corporation_dataexporter.has_data,
    }
    context = add_info_to_context(request, context)

    return render(
        request,
        "ledger/corpledger/admin/corporation_administration.html",
        context=context,
    )


@login_required
@permission_required("ledger.manage_access")
@require_POST
def corporation_delete(request, corporation_id):
    """
    Character Delete
    """
    perms = get_manage_corporation(request, corporation_id)[0]

    if perms is False:
        msg = _("Permission Denied")
        return JsonResponse(
            {"success": False, "message": msg}, status=HTTPStatus.FORBIDDEN, safe=False
        )
    if perms is None:
        msg = _("Corporation not found")
        return JsonResponse(
            {"success": False, "message": msg}, status=HTTPStatus.NOT_FOUND, safe=False
        )

    audit = CorporationAudit.objects.get(corporation__corporation_id=corporation_id)
    audit.delete()

    msg = _(f"{audit.corporation.corporation_name} successfully deleted")
    return JsonResponse(
        {"success": True, "message": msg}, status=HTTPStatus.OK, safe=False
    )
